from collections import Counter
import typing

from karp.lex.application.repositories import (
    ResourceUnitOfWork,
    EntryUowRepositoryUnitOfWork,
)
from karp.lex.application.queries import (
    GetEntryRepositoryId,
)
from karp.search.application.queries import (
    EntryDto,
    QueryRequest,
    QueryResponse,
    QuerySplitResponse,
    SearchService,
    StatisticsDto,
)
from karp.search.domain import query_dsl


class GenericSearchService(SearchService):
    def __init__(
        self,
        get_entry_repo_id: GetEntryRepositoryId,
        entry_uow_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> None:
        self.get_entry_repo_id = get_entry_repo_id
        self.entry_uow_repo_uow = entry_uow_repo_uow
        self.parser = query_dsl.KarpQueryV6Parser(
            semantics=query_dsl.KarpQueryV6ModelBuilderSemantics()
        )

    def statistics(
        self, resource_id: str, field: str
    ) -> typing.Iterable[StatisticsDto]:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_uow_repo_uow, self.entry_uow_repo_uow.repo.get_by_id(
            entry_repo_id
        ) as uw:
            all_entries = uw.repo.all_entries()
            values_of_field = (entry.body.get(field) for entry in all_entries)
            aggregate = Counter(values_of_field)
            return (
                StatisticsDto(value=key, count=count)
                for key, count in aggregate.items()
                if key is not None
            )

    def query(self, request: QueryRequest):
        if request.q:
            model = self.parser.parse(request.q)
        resource_id = request.resource_ids[0]
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_uow_repo_uow, self.entry_uow_repo_uow.repo.get_by_id(
            entry_repo_id
        ) as uw:
            all_entries = uw.repo.all_entries()
            hits = [
                EntryDto(
                    id=entry.entry_id,
                    version=entry.version,
                    last_modified=entry.last_modified,
                    last_modified_by=entry.last_modified_by,
                    resource=resource_id,
                    entry=entry.body,
                )
                for entry in all_entries
            ]

        if request.sort:
            field, order = request.sort[0].split("|")
            hits.sort(
                key=lambda entry: entry.entry.get(field, 0),
                reverse=order.lower() == "desc",
            )
        response = QueryResponse(
            hits=hits,
            total=len(hits),
        )
        if request.lexicon_stats:
            response.distribution = {resource_id: len(hits)}

        return response

    def query_split(self, request: QueryRequest):
        if request.q:
            model = self.parser.parse(request.q)
        dist = {}
        for resource_id in request.resource_ids:
            entry_repo_id = self.get_entry_repo_id.query(resource_id)
            with self.entry_uow_repo_uow, self.entry_uow_repo_uow.repo.get_by_id(
                entry_repo_id
            ) as uw:
                all_entries = uw.repo.all_entries()
                dist[resource_id] = [
                    EntryDto(
                        id=entry.entry_id,
                        version=entry.version,
                        last_modified=entry.last_modified,
                        last_modified_by=entry.last_modified_by,
                        resource=resource_id,
                        entry=entry.body,
                    )
                    for entry in all_entries
                ]

        return QuerySplitResponse(
            hits=[],
            total=0,
            distribution={resource: len(hits) for resource, hits in dist.items()},
        )

    def search_ids(self):
        pass



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.parsing import leftrec, nomemo, isname  # noqa
from tatsu.infos import ParserConfig
from tatsu.util import re, generic_main  # noqa


KEYWORDS = {}  # type: ignore


class KarpQueryV6Buffer(Buffer):
    def __init__(self, text, /, config: ParserConfig = None, **settings):
        base_config = ParserConfig.new(
            owner=self,
            whitespace=None,
            nameguard=None,
            comments_re=None,
            eol_comments_re=None,
            ignorecase=False,
            namechars="",
            parseinfo=False,
        )
        config = base_config.replace_config(config)
        config = config.merge(**settings)
        super().__init__(text, config=config)


class KarpQueryV6Parser(Parser):
    def __init__(self, config: ParserConfig = None, **settings):
        base_config = ParserConfig.new(
            owner=self,
            whitespace=None,
            nameguard=None,
            comments_re=None,
            eol_comments_re=None,
            ignorecase=False,
            namechars="",
            parseinfo=False,
            keywords=KEYWORDS,
        )
        config = base_config.replace_config(config)
        config = config.merge(**settings)
        super().__init__(config=config)

    @tatsumasu()
    def _start_(self):  # noqa
        self._expression_()
        self._check_eof()

    @tatsumasu()
    def _expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._logical_expression_()
            with self._option():
                self._query_expression_()
            self._error(
                "expecting one of: "
                "<and> <or> <not> <logical_expression>"
                "<contains> <endswith> <equals> <exists>"
                "<freergxp> <freetext> <gt> <gte> <lt>"
                "<lte> <missing> <regexp> <startswith>"
                "<query_expression>"
            )

    @tatsumasu()
    def _query_expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._contains_()
            with self._option():
                self._endswith_()
            with self._option():
                self._equals_()
            with self._option():
                self._exists_()
            with self._option():
                self._freergxp_()
            with self._option():
                self._freetext_()
            with self._option():
                self._gt_()
            with self._option():
                self._gte_()
            with self._option():
                self._lt_()
            with self._option():
                self._lte_()
            with self._option():
                self._missing_()
            with self._option():
                self._regexp_()
            with self._option():
                self._startswith_()
            self._error(
                "expecting one of: "
                "'contains' 'endswith' 'equals' 'exists'"
                "'freergxp' <freetext_any>"
                "<freetext_string> <freetext> 'gt' 'gte'"
                "'lt' 'lte' 'missing' 'regexp'"
                "'startswith'"
            )

    @tatsumasu()
    def _logical_expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._and_()
            with self._option():
                self._or_()
            with self._option():
                self._not_()
            self._error("expecting one of: " "'and' 'or' 'not'")

    @tatsumasu("And")
    def _and_(self):  # noqa
        self._token("and")
        self.name_last_node("op")
        self._token("(")
        self._expression_()
        self.add_last_node_to_name("exps")

        def block2():
            self._token("||")
            self._expression_()
            self.add_last_node_to_name("exps")

        self._closure(block2)
        self._token(")")
        self._define(["op"], ["exps"])

    @tatsumasu("Contains")
    def _contains_(self):  # noqa
        self._token("contains")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._string_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Endswith")
    def _endswith_(self):  # noqa
        self._token("endswith")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._string_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Equals")
    def _equals_(self):  # noqa
        self._token("equals")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._argument_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Exists")
    def _exists_(self):  # noqa
        self._token("exists")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._define(["field", "op"], [])

    @tatsumasu("Freergxp")
    def _freergxp_(self):  # noqa
        self._token("freergxp")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("arg")
        self._define(["arg", "op"], [])

    @tatsumasu()
    def _freetext_(self):  # noqa
        with self._choice():
            with self._option():
                self._freetext_any_()
            with self._option():
                self._freetext_string_()
            self._error(
                "expecting one of: " "'freetext' <freetext_any>" "<freetext_string>"
            )

    @tatsumasu("FreetextAnyButString")
    def _freetext_any_(self):  # noqa
        self._token("freetext")
        self.name_last_node("op")
        self._token("|")
        self._any_but_string_()
        self.name_last_node("arg")
        self._define(["arg", "op"], [])

    @tatsumasu("FreetextString")
    def _freetext_string_(self):  # noqa
        self._token("freetext")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("arg")
        self._define(["arg", "op"], [])

    @tatsumasu("Gt")
    def _gt_(self):  # noqa
        self._token("gt")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._argument_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Gte")
    def _gte_(self):  # noqa
        self._token("gte")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._argument_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Lt")
    def _lt_(self):  # noqa
        self._token("lt")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._argument_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Lte")
    def _lte_(self):  # noqa
        self._token("lte")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._argument_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Missing")
    def _missing_(self):  # noqa
        self._token("missing")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._define(["field", "op"], [])

    @tatsumasu("Not")
    def _not_(self):  # noqa
        self._token("not")
        self.name_last_node("op")
        self._token("(")
        self._expression_()
        self.add_last_node_to_name("exps")

        def block2():
            self._token("||")
            self._expression_()
            self.add_last_node_to_name("exps")

        self._closure(block2)
        self._token(")")
        self._define(["op"], ["exps"])

    @tatsumasu("Or")
    def _or_(self):  # noqa
        self._token("or")
        self.name_last_node("op")
        self._token("(")
        self._expression_()
        self.add_last_node_to_name("exps")

        def block2():
            self._token("||")
            self._expression_()
            self.add_last_node_to_name("exps")

        self._closure(block2)
        self._token(")")
        self._define(["op"], ["exps"])

    @tatsumasu("Regexp")
    def _regexp_(self):  # noqa
        self._token("regexp")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._string_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu("Startswith")
    def _startswith_(self):  # noqa
        self._token("startswith")
        self.name_last_node("op")
        self._token("|")
        self._string_()
        self.name_last_node("field")
        self._token("|")
        self._string_()
        self.name_last_node("arg")
        self._define(["arg", "field", "op"], [])

    @tatsumasu()
    def _argument_(self):  # noqa
        with self._choice():
            with self._option():
                self._integer_()
            with self._option():
                self._string_()
            self._error("expecting one of: " "\\d+ <integer> [^|()]+ <string>")

    @tatsumasu()
    def _any_but_string_(self):  # noqa
        self._integer_()

    @tatsumasu()
    def _string_(self):  # noqa
        self._pattern("[^|()]+")

    @tatsumasu("int")
    def _integer_(self):  # noqa
        self._pattern("\\d+")


class KarpQueryV6Semantics(object):
    def start(self, ast):  # noqa
        return ast

    def expression(self, ast):  # noqa
        return ast

    def query_expression(self, ast):  # noqa
        return ast

    def logical_expression(self, ast):  # noqa
        return ast

    def and_(self, ast):  # noqa
        return ast

    def contains(self, ast):  # noqa
        return ast

    def endswith(self, ast):  # noqa
        return ast

    def equals(self, ast):  # noqa
        return ast

    def exists(self, ast):  # noqa
        return ast

    def freergxp(self, ast):  # noqa
        return ast

    def freetext(self, ast):  # noqa
        return ast

    def freetext_any(self, ast):  # noqa
        return ast

    def freetext_string(self, ast):  # noqa
        return ast

    def gt(self, ast):  # noqa
        return ast

    def gte(self, ast):  # noqa
        return ast

    def lt(self, ast):  # noqa
        return ast

    def lte(self, ast):  # noqa
        return ast

    def missing(self, ast):  # noqa
        return ast

    def not_(self, ast):  # noqa
        return ast

    def or_(self, ast):  # noqa
        return ast

    def regexp(self, ast):  # noqa
        return ast

    def startswith(self, ast):  # noqa
        return ast

    def argument(self, ast):  # noqa
        return ast

    def any_but_string(self, ast):  # noqa
        return ast

    def string(self, ast):  # noqa
        return ast

    def integer(self, ast):  # noqa
        return ast


def main(filename, start=None, **kwargs):
    if start is None:
        start = "start"
    if not filename or filename == "-":
        text = sys.stdin.read()
    else:
        with open(filename) as f:
            text = f.read()
    parser = KarpQueryV6Parser()
    return parser.parse(text, rule_name=start, filename=filename, **kwargs)


if __name__ == "__main__":
    import json
    from tatsu.util import asjson

    ast = generic_main(main, KarpQueryV6Parser, name="KarpQueryV6")
    data = asjson(ast)
    print(json.dumps(data, indent=2))



import logging
import os
from typing import Union

from karp.db_infrastructure import Database

from karp.main.config import DATABASE_URL, DatabaseUrl


logger = logging.getLogger(__name__)


def connect_to_db(db_url: Union[str, DatabaseUrl]) -> Database:
    try:
        return Database(db_url)
    except Exception as e:
        logger.warn("--- DB CONNECTION ERROR ---")
        logger.warn(e)
        logger.warn("--- DB CONNECTION ERROR ---")


def close_db_connection(database: Database) -> None:
    try:
        database.disconnect()
    except Exception as e:
        logger.warn("--- DB DISCONNECT ERROR ---")
        logger.warn(e)
        logger.warn("--- DB DISCONNECT ERROR ---")



from typing import Iterable, Optional

from sqlalchemy.sql import select

from karp.lex import ListEntryRepos, EntryRepoDto, ReadOnlyEntryRepoRepositry
from karp.lex_infrastructure.queries.base import SqlQuery
from karp.lex_infrastructure.sql.sql_models import EntryUowModel


class SqlListEntryRepos(ListEntryRepos, SqlQuery):
    def query(self) -> Iterable[EntryRepoDto]:
        stmt = select(EntryUowModel)
        return (_row_to_dto(row) for row in self._conn.execute(stmt))


def _row_to_dto(row_proxy) -> EntryRepoDto:
    return EntryRepoDto(
        name=row_proxy.name,
        entity_id=row_proxy.entity_id,
        repository_type=row_proxy.type,
    )


class SqlReadOnlyEntryRepoRepositry(SqlQuery, ReadOnlyEntryRepoRepositry):
    def get_by_name(self, name: str) -> Optional[EntryRepoDto]:
        stmt = select(EntryUowModel).filter_by(name=name)

        row = self._conn.execute(stmt).first()

        return _row_to_dto(row) if row else None





class QueryDSLError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ParseError(QueryDSLError):
    def __init__(self, message: str) -> None:
        super().__init__(message)

    def __repr__(self) -> str:
        return "ParseError message='{}'".format(self.message)


class SyntaxError(ParseError):
    def __init__(self, message: str):
        super().__init__(message)

    def __repr__(self) -> str:
        return "SyntaxError message='{}'".format(self.message)



class AuthError(Exception):
    """Auth base exception."""


class ResourceNotFound(AuthError):
    pass


class TokenError(AuthError):
    """General token error."""


class ExpiredToken(TokenError):
    """The given token has expired."""


class InvalidTokenSignature(TokenError):
    pass


class InvalidTokenAudience(TokenError):
    pass


class InvalidTokenPayload(TokenError):
    pass



import json
from logging.config import dictConfig
import logging
import os
import typing
from dataclasses import dataclass


try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points  # type: ignore

import injector
from json_streams import jsonlib
from sqlalchemy import pool
from sqlalchemy.engine import Engine, create_engine, url as sa_url
import logging
import asgi_correlation_id

from karp.foundation.environs_sqlalchemyurl import sqlalchemy_url
from karp.lex import Lex
from karp.lex_infrastructure import GenericLexInfrastructure, LexInfrastructure
from karp.search_infrastructure import (
    GenericSearchInfrastructure,
    Es6SearchIndexMod,
    GenericSearchIndexMod,
    SearchInfrastructure,
)
from karp.main import config, modules
from karp.main.modules import (
    CommandBusMod,
    Db,
    EventBusMod,
    ElasticSearchMod,
    install_auth_service,
)
from karp.search import Search


@dataclass
class AppContext:
    container: injector.Injector
    settings: typing.Dict


def bootstrap_app(container=None) -> AppContext:
    env = config.load_env()
    db_url = config.parse_database_url(env)
    es_enabled = env.bool("ELASTICSEARCH_ENABLED", False)
    if es_enabled:
        es_url = env("ELASTICSEARCH_HOST")
    else:
        es_url = env("ELASTICSEARCH_HOST", "")
    settings = {
        "auth.jwt.pubkey.path": env("AUTH_JWT_PUBKEY_PATH", None),
        "auth.name": env("AUTH_NAME", ""),
        "tracking.matomo.idsite": env("TRACKING_MATOMO_IDSITE", None),
        "tracking.matomo.url": env("TRACKING_MATOMO_URL", None),
        "tracking.matomo.token": env("TRACKING_MATOMO_TOKEN", None),
        "es.index_prefix": env("ES_INDEX_PREFIX", None),
    }
    # if n ot container:
    # container = AppContainer()
    # container.config.core.logging.from_value(_logging_config())
    # container.config.db.url.from_value(config.DB_URL)
    # container.config.search_service.type.from_env(
    # "SEARCH_CONTEXT", "sql_search_service"
    # )
    # bus = container.bus()
    # bus.handle(events.AppStarted())  # needed? ?
    configure_logging(settings=settings)
    # logging.config.dictConfig(_logging_config())  # type: ignore
    search_service = env("SEARCH_CONTEXT", "sql_search_service")
    engine = _create_db_engine(db_url)
    dependency_injector = _setup_dependency_injection(engine, es_url=es_url)
    _setup_search_context(dependency_injector, search_service, settings=settings)
    return AppContext(dependency_injector, settings)


def _create_db_engine(db_url: config.DatabaseUrl) -> Engine:
    kwargs = {}
    if str(db_url).startswith("sqlite"):
        kwargs["poolclass"] = pool.SingletonThreadPool
        kwargs["connect_args"] = {"check_same_thread": False}
    engine_echo = False
    return create_engine(db_url, echo=engine_echo, future=True, **kwargs)


def _setup_dependency_injection(
    engine: Engine,
    es_url: str,
) -> injector.Injector:
    return injector.Injector(
        [
            Db(engine),
            CommandBusMod(),
            EventBusMod(),
            ElasticSearchMod(es_url),
            Lex(),
            LexInfrastructure(),
            GenericLexInfrastructure(),
            Search(),
            GenericSearchInfrastructure(),
            SearchInfrastructure(),
            # SearchServiceMod(search_service),
        ],
        auto_bind=False,
    )


def _setup_search_context(
    container: injector.Injector, search_service: str, settings: dict
) -> None:
    if search_service.lower() == "es6_search_service":
        container.binder.install(
            Es6SearchIndexMod(index_prefix=settings.get("es.index_prefix"))
        )
    else:
        container.binder.install(GenericSearchIndexMod())


def configure_logging(settings: dict[str, str]) -> None:
    dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "filters": {
                "correlation_id": {
                    "()": asgi_correlation_id.CorrelationIdFilter,
                    "uuid_length": 32,
                }
            },
            "formatters": {
                "console": {
                    "class": "logging.Formatter",
                    # 'datefmt':  '%H:%M:%S',
                    "format": "%(levelname)s:\t\b%(asctime)s %(name)s:%(lineno)d [%(correlation_id)s] %(message)s",
                },
                "json": {
                    "class": "pythonjsonlogger.jsonlogger.JsonFormatter",
                    # 'format': '%(message)s',
                    "format": "%(asctime)s %(levelname)s %(name)s %(process)d %(funcName)s %(lineno)d %(message)s",
                },
                "standard": {
                    "format": "%(asctime)s-%(levelname)s-%(name)s-%(process)d::%(module)s|%(lineno)s:: %(message)s",
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "filters": ["correlation_id"],
                    "formatter": "console",
                    # "level": 'DEBUG',
                    "stream": "ext://sys.stderr",
                },
                "json": {
                    "class": "logging.StreamHandler",
                    "filters": ["correlation_id"],
                    "formatter": "json",
                },
                # "email": {
                #     "class": "logging.handlers.SMTPHandler",
                #     "mailhost": "localhost",
                #     "formatter": "standard",
                #     "level": "WARNING",
                #     "fromaddr": config.LOG_MAIL_FROM,
                #     "toaddrs": config.LOG_MAIL_TOS,
                #     "subject": "Error in Karp backend!",
                # },
            },
            "loggers": {
                "karp": {
                    "handlers": ["json"],  # ["console", "email"],
                    "level": "DEBUG",  # config.CONSOLE_LOG_LEVEL,
                    "propagate": True,
                },
                # third-party package loggers
                "sqlalchemy": {"handlers": ["json"], "level": "WARNING"},
                "uvicorn.access": {"handlers": ["json"], "level": "INFO"},
            },
        }
    )


# def bootstrap(
#    resource_uow: unit_of_work.ResourceUnitOfWork = None,
#    entry_uows: unit_of_work.EntriesUnitOfWork = None,
#    authservice: auth_service.AuthService = None,
#    index_uow: unit_of_work.IndexUnitOfWork = None,
#    entry_uow_factory: unit_of_work.EntryUowFactory = None,
#    raise_on_all_errors: bool = False,
# ) -> messagebus.MessageBus:
#    setup_logging()
#    load_infrastructure()
#    if authservice is None:
#        authservice = auth_service.AuthService.create(config.AUTH_CONTEXT)
#    if resource_uow is None:
#        resource_uow = sql_unit_of_work.SqlResourceUnitOfWork()
#    if entry_uows is None:
#        entry_uows = unit_of_work.EntriesUnitOfWork()
#    if entry_uow_factory is None:
#        entry_uow_factory = unit_of_work.DefaultEntryUowFactory()
#    if index_uow is None:
#        index_uow = unit_of_work.IndexUnitOfWork.create(config.SEARCH_CONTEXT)
#    bus = messagebus.MessageBus(
#        resource_uow=resource_uow,
#        entry_uows=entry_uows,
#        # auth_service=authservice,
#        search_service_uow=index_uow,
#        entry_uow_factory=entry_uow_factory,
#        raise_on_all_errors=raise_on_all_errors,
#    )
#    bus.handle(events.AppStarted())  # needed? ?
#    return bus


def setup_logging():

    # Clear Gunicorn access log to remove duplicate requests logging
    # logging.getLogger("gunicorn.access").handlers.clear()
    logging.basicConfig(
        format="%(asctime)s %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        level=logging.INFO,
    )
    logger = logging.getLogger("karp")
    logger.setLevel(logging.INFO)
    # structlog.configure(
    #     processors=[
    #         structlog.stdlib.add_log_level,
    #         structlog.threadlocal.merge_threadlocal,
    #         structlog.processors.TimeStamper(fmt="iso"),
    #         structlog.processors.JSONRenderer(
    #             serializer=json.dumps, sort_keys=True),
    #     ],
    #     wrapper_class=structlog.stdlib.BoundLogger,
    #     context_class=dict,
    #     cache_logger_on_first_use=True,
    #     logger_factory=structlog.stdlib.LoggerFactory(),
    # )
    logger = logging.getLogger(__name__)
    # if app.config.get("LOG_TO_SLACK"):
    #     slack_handler = slack_logging.get_slack_logging_handler(
    #         app.config.get("SLACK_SECRET")
    #     )
    #     logger.addHandler(slack_handler)
    # console_handler = logging.StreamHandler()
    # formatter = logging.Formatter(
    #     "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    # )
    # console_handler.setFormatter(formatter)
    # logger.setLevel(logging.DEBUG)
    # logger.addHandler(console_handler)
    return logger


def load_infrastructure():
    modules.load_modules("karp.infrastructure")



from sqlalchemy.engine.url import URL, make_url


def sqlalchemy_url(
    db_url: str,
    driver: str = None,
    host: str = None,
    port: int = None,
    user: str = None,
    password: str = None,
    database: str = None,
) -> URL:
    if db_url:
        return make_url(db_url)
    return URL.create(
        drivername=driver,
        username=user,
        password=password,
        host=host,
        port=port,
        database=database,
    )



import functools
import logging
import time

import typer

from karp.main.errors import KarpError

logger = logging.getLogger("karp")


def cli_error_handler(func):
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KarpError as e:
            logger.error(e.message)
            raise typer.Exit(e.code)

    return func_wrapper


def cli_timer(func):
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        before_t = time.time()
        result = func(*args, **kwargs)
        typer.echo("Command took: %0.1fs" % (time.time() - before_t))
        return result

    return func_wrapper



from unittest import mock

import pytest
import ulid

from karp.foundation.events import EventBus
from karp import lex
from karp.lex_infrastructure import SqlEntryUowV1Creator, SqlEntryUowV2Creator
from karp.tests.unit.lex import factories


@pytest.fixture
def example_uow() -> lex.CreateEntryRepository:
    return factories.CreateEntryRepositoryFactory()


@pytest.fixture
def sql_entry_uow_v1_creator(sqlite_session_factory) -> SqlEntryUowV1Creator:
    return SqlEntryUowV1Creator(
        event_bus=mock.Mock(spec=EventBus),
        session_factory=sqlite_session_factory,
    )


@pytest.fixture
def sql_entry_uow_v2_creator(sqlite_session_factory) -> SqlEntryUowV2Creator:
    return SqlEntryUowV2Creator(
        event_bus=mock.Mock(spec=EventBus),
        session_factory=sqlite_session_factory,
    )


class TestSqlEntryUowV1:
    def test_creator_repository_type(
        self,
        sql_entry_uow_v1_creator: SqlEntryUowV1Creator,
    ):
        assert sql_entry_uow_v1_creator.repository_type == "sql_entries_v1"

    def test_uow_repository_type(
        self,
        sql_entry_uow_v1_creator: SqlEntryUowV1Creator,
        example_uow: lex.CreateEntryRepository,
    ):
        entry_uow = sql_entry_uow_v1_creator(
            **example_uow.dict(exclude={"repository_type"})
        )
        assert entry_uow.repository_type == "sql_entries_v1"

    def test_repo_table_name(
        self,
        sql_entry_uow_v1_creator: SqlEntryUowV1Creator,
        example_uow: lex.CreateEntryRepository,
    ):
        entry_uow = sql_entry_uow_v1_creator(
            **example_uow.dict(exclude={"repository_type"})
        )
        with entry_uow as uw:
            assert uw.repo.history_model.__tablename__ == example_uow.name


class TestSqlEntryUowV2:
    def test_creator_repository_type(
        self,
        sql_entry_uow_v2_creator: SqlEntryUowV2Creator,
    ):
        assert sql_entry_uow_v2_creator.repository_type == "sql_entries_v2"

    def test_uow_repository_type(
        self,
        sql_entry_uow_v2_creator: SqlEntryUowV2Creator,
        example_uow: lex.CreateEntryRepository,
    ):
        entry_uow = sql_entry_uow_v2_creator(
            **example_uow.dict(exclude={"repository_type"})
        )
        assert entry_uow.repository_type == "sql_entries_v2"

    def test_repo_table_name(
        self,
        sql_entry_uow_v2_creator: SqlEntryUowV2Creator,
        example_uow: lex.CreateEntryRepository,
    ):
        entry_uow = sql_entry_uow_v2_creator(
            **example_uow.dict(exclude={"repository_type"})
        )
        random_part = ulid.from_uuid(entry_uow.entity_id).randomness().str
        with entry_uow as uw:
            assert (
                uw.repo.history_model.__tablename__
                == f"{example_uow.name}_{random_part}"
            )



"""Base class for SQL repositories."""
import logging

from sqlalchemy.orm import Session

logger = logging.getLogger("karp")


class SqlRepository:
    def __init__(self, session: Session) -> None:
        self._session: Session = session

    def set_session(self, session):
        self._session = session

    def unset_session(self):
        self._session = None

    def _check_has_session(self):
        if self._session is None:
            raise RuntimeError("No session, use with unit_of_work.")



import abc
import ulid
from typing import Generic, TypeVar, Optional, Type, Union

from karp.foundation.value_objects import unique_id
from .errors import NotFoundError

EntityType = TypeVar("EntityType")


class Repository(Generic[EntityType], abc.ABC):
    EntityNotFound: Type[NotFoundError] = NotFoundError

    def __init__(self):
        self.seen = set()

    def save(self, entity: EntityType):
        self._save(entity)
        self.seen.add(entity)

    @abc.abstractmethod
    def _save(self, entity: EntityType):
        raise NotImplementedError()

    def _check_id_has_correct_type(self, id_) -> None:
        if not isinstance(id_, unique_id.UniqueId):
            raise ValueError(f"expected UniqueId, got '{id_}' (type: ´{type(id_)}')")

    def _ensure_correct_id_type(self, v) -> unique_id.UniqueId:
        try:
            return unique_id.parse(v)
        except ValueError as exc:
            raise ValueError(
                f"expected valid UniqueId, got '{v}' (type: ´{type(v)}')"
            ) from exc

    def by_id(
        self, id_: unique_id.UniqueId, *, version: Optional[int] = None, **kwargs
    ) -> EntityType:
        # self._check_id_has_correct_type(id_)
        entity = self._by_id(self._ensure_correct_id_type(id_), version=version)
        if entity:
            self.seen.add(entity)
        else:
            raise self.EntityNotFound(f"Entity with id={id_} is not found")
        return entity

    get_by_id = by_id

    def get_by_id_optional(
        self, id_: unique_id.UniqueId, *, version: Optional[int] = None, **kwargs
    ) -> Optional[EntityType]:
        # self._check_id_has_correct_type(id_)
        entity = self._by_id(self._ensure_correct_id_type(id_), version=version)
        if entity:
            self.seen.add(entity)
        return entity

    @abc.abstractmethod
    def _by_id(
        self,
        id_: Union[unique_id.UniqueId, ulid.ULID, str],
        *,
        version: Optional[int] = None,
        **kwargs,
    ) -> Optional[EntityType]:
        raise NotImplementedError()

    @abc.abstractmethod
    def num_entities(self) -> int:
        ...



import abc
from typing import Iterable, Optional, Dict

import pydantic

from karp.foundation.value_objects import UniqueIdStr, UniqueId
from karp.foundation.value_objects.unique_id import UniqueIdPrimitive


class ResourceDto(pydantic.BaseModel):
    resource_id: str
    entity_id: UniqueIdStr
    is_published: bool
    version: int
    name: str
    last_modified_by: str
    message: str
    last_modified: float
    config: Dict
    entry_repository_id: UniqueIdStr
    discarded: bool


class GetPublishedResources(abc.ABC):
    @abc.abstractmethod
    def query(self) -> Iterable[ResourceDto]:
        pass


class GetResources(abc.ABC):
    @abc.abstractmethod
    def query(self) -> Iterable[ResourceDto]:
        pass


class GetEntryRepositoryId(abc.ABC):
    @abc.abstractmethod
    def query(self, resource_id: str) -> UniqueId:
        raise NotImplementedError()


class ReadOnlyResourceRepository(abc.ABC):
    def get_by_resource_id(
        self, resource_id: str, version: Optional[int] = None
    ) -> Optional[ResourceDto]:
        resource = self._get_by_resource_id(resource_id)
        if not resource:
            return None

        if version is not None:
            resource = self.get_by_id(resource.entity_id, version=version)
        return resource

    @abc.abstractmethod
    def get_by_id(
        self, entity_id: UniqueIdPrimitive, version: Optional[int] = None
    ) -> Optional[ResourceDto]:
        pass

    @abc.abstractmethod
    def _get_by_resource_id(self, resource_id: str) -> Optional[ResourceDto]:
        pass

    @abc.abstractmethod
    def get_published_resources(self) -> Iterable[ResourceDto]:
        pass



import abc
import logging
import typing
import uuid
from typing import Dict, List, Optional, Tuple, Union

from karp.foundation import events, repository, unit_of_work
from karp.lex.domain import entities

from karp.lex.domain import errors

logger = logging.getLogger("karp")


class ResourceRepository(repository.Repository[entities.Resource]):
    EntityNotFound = errors.ResourceNotFound

    @abc.abstractmethod
    def check_status(self):
        pass

    @abc.abstractmethod
    def resource_ids(self) -> typing.Iterable[str]:
        raise NotImplementedError()

    def get_by_resource_id(
        self, resource_id: str, *, version: Optional[int] = None
    ) -> entities.Resource:
        resource = self.get_by_resource_id_optional(resource_id, version=version)

        if not resource:
            raise self.EntityNotFound(
                f"Entity with resource_id='{resource_id}' can't be found."
            )

        return resource

    def get_by_resource_id_optional(
        self, resource_id: str, *, version: Optional[int] = None
    ) -> typing.Optional[entities.Resource]:
        resource = self._by_resource_id(resource_id)
        if not resource:
            return None

        if version:
            resource = self._by_id(resource.entity_id, version=version)
        if resource:
            self.seen.add(resource)
        return resource

    by_resource_id = get_by_resource_id

    @abc.abstractmethod
    def _by_resource_id(
        self,
        resource_id: str,
    ) -> Optional[entities.Resource]:
        raise NotImplementedError()

    # @abc.abstractmethod
    # def resources_with_id(self, resource_id: str):
    #     raise NotImplementedError()

    # @abc.abstractmethod
    # def resource_with_id_and_version(self, resource_id: str, version: int):
    #     raise NotImplementedError()

    # @abc.abstractmethod
    # def get_active_resource(self, resource_id: str) -> Optional[Resource]:
    #     raise NotImplementedError()

    def get_published_resources(self) -> typing.List[entities.Resource]:
        published_resources = []
        for resource in self._get_published_resources():
            self.seen.add(resource)
            published_resources.append(resource)
        return published_resources

    @abc.abstractmethod
    def _get_published_resources(self) -> typing.Iterable[entities.Resource]:
        raise NotImplementedError()

    def get_all_resources(self) -> typing.List[entities.Resource]:
        all_resources = []
        for resource in self._get_all_resources():
            self.seen.add(resource)
            all_resources.append(resource)
        return all_resources

    @abc.abstractmethod
    def _get_all_resources(self) -> typing.Iterable[entities.Resource]:
        raise NotImplementedError()


class ResourceUnitOfWork(unit_of_work.UnitOfWork[ResourceRepository]):
    def __init__(self, event_bus: events.EventBus):
        unit_of_work.UnitOfWork.__init__(self, event_bus=event_bus)

    @property
    def resources(self) -> ResourceRepository:
        return self.repo



import collections
import logging
import typing

# from karp.infrastructure.unit_of_work import unit_of_work
from typing import Any, Dict, Iterator, List, Optional, Tuple
from karp.foundation.value_objects.unique_id import UniqueId
from karp.lex.application.queries.entries import EntryDto

from karp.lex.domain import entities
from karp.lex.domain.entities.entry import Entry
from karp.lex.application.queries import (
    GetReferencedEntries,
    ReferenceDto,
    ReadOnlyResourceRepository,
)
from karp.lex.application.repositories import EntryUowRepositoryUnitOfWork

# from karp.resourcemgr import get_resource
# import karp.resourcemgr.entryread as entryread
from karp.lex.domain.errors import EntryNotFound

# from karp.services import context

logger = logging.getLogger(__name__)


class GenericGetReferencedEntries(GetReferencedEntries):
    def __init__(
        self,
        resource_repo: ReadOnlyResourceRepository,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> None:
        super().__init__()
        self.resource_repo = resource_repo
        self.entry_repo_uow = entry_repo_uow

    def query(
        self,
        resource_id: str,
        # version: typing.Optional[int],
        entry_id: UniqueId,
    ) -> typing.Iterable[ReferenceDto]:
        logger.debug(f"network.get_referenced_entries: resource_id={resource_id}")
        logger.debug(f"network.get_referenced_entries: entry_id={entry_id}")
        resource_refs, resource_backrefs = self.get_refs(
            resource_id,  # version=version
        )

        resource = self.resource_repo.get_by_resource_id(resource_id)
        if not resource:
            raise RuntimeError(f"resource '{resource_id}' not found")
        with self.entry_repo_uow:
            entry_uow = self.entry_repo_uow.repo.get_by_id(resource.entry_repository_id)
        with entry_uow as uw:
            src_entry = uw.repo.by_id(
                entry_id,  # version=version
            )
            # if not src_entry:
            #     raise EntryNotFoundError(
            #         resource.resource_id, entry_id, resource_version=resource.version
            #     )

        for (
            ref_resource_id,
            ref_resource_version,
            field_name,
            field,
        ) in resource_backrefs:
            other_resource = self.resource_repo.get_by_resource_id(
                ref_resource_id  # , version=version
            )
            if not other_resource:
                logger.info("resource %s not found, skipping ...", ref_resource_id)
                continue
            with self.entry_repo_uow:
                other_entry_uow = self.entry_repo_uow.repo.get_by_id(
                    other_resource.entry_repository_id
                )
            with other_entry_uow as entries_uw:
                for entry in entries_uw.repo.by_referenceable({field_name: entry_id}):
                    yield _create_ref(ref_resource_id, ref_resource_version, entry)

        # src_body = json.loads(src_entry.body)
        for (ref_resource_id, ref_resource_version, field_name, field) in resource_refs:
            ids = src_entry.body.get(field_name)
            if ids is None:
                continue
            if not field.get("collection", False):
                ids = [ids]
            if ref_resource := self.resource_repo.get_by_resource_id(ref_resource_id):
                with self.entry_repo_uow.repo.get_by_id(
                    ref_resource.entry_repository_id
                ) as entries_uw:
                    for ref_entry_id in ids:
                        if entry := entries_uw.repo.get_by_id_optional(
                            ref_entry_id, version=ref_resource_version
                        ):
                            yield _create_ref(
                                ref_resource_id, ref_resource_version, entry
                            )

    def get_refs(
        self, resource_id, version=None
    ) -> Tuple[List[Tuple[str, int, str, Dict]], List[Tuple[str, int, str, Dict]]]:
        """
        Goes through all other resource configs finding resources and fields that refer to this resource
        """
        resource_backrefs = collections.defaultdict(
            lambda: collections.defaultdict(dict)
        )
        resource_refs = collections.defaultdict(lambda: collections.defaultdict(dict))

        src_resource = self.resource_repo.get_by_resource_id(
            resource_id, version=version
        )

        if not src_resource:
            raise RuntimeError(f"Can't find resource '{resource_id}'")
        all_other_resources = [
            resource
            for resource in self.resource_repo.get_published_resources()
            if resource.resource_id != resource_id
        ]

        for field_name, field in src_resource.config["fields"].items():
            if "ref" in field:
                if "resource_id" not in field["ref"]:
                    resource_backrefs[resource_id][version][field_name] = field
                    resource_refs[resource_id][version][field_name] = field
                else:
                    resource_refs[field["ref"]["resource_id"]][
                        field["ref"]["resource_version"]
                    ][field_name] = field
            elif "function" in field and "multi_ref" in field["function"]:
                virtual_field = field["function"]["multi_ref"]
                ref_field = virtual_field["field"]
                if "resource_id" in virtual_field:
                    resource_backrefs[virtual_field["resource_id"]][
                        virtual_field["resource_version"]
                    ][ref_field] = None
                else:
                    resource_backrefs[resource_id][version][ref_field] = None

        for other_resource in all_other_resources:
            # other_resource = get_resource(
            #     resource_def.resource_id, version=resource_def.version
            # )
            for field_name, field in other_resource.config["fields"].items():
                ref = field.get("ref")
                if (
                    ref
                    and ref.get("resource_id") == resource_id
                    and ref.get("resource_version") == version
                ):
                    resource_backrefs[other_resource.resource_id][
                        other_resource.version
                    ][field_name] = field

        def flatten_dict(ref_dict):
            ref_list = []
            for ref_resource_id, versions in ref_dict.items():
                for ref_version, field_names in versions.items():
                    for field_name, field in field_names.items():
                        ref_list.append(
                            (ref_resource_id, ref_version, field_name, field)
                        )
            return ref_list

        return flatten_dict(resource_refs), flatten_dict(resource_backrefs)


def _create_ref(resource_id: str, resource_version: int, entry: Entry) -> ReferenceDto:
    return ReferenceDto(
        resource_id=resource_id,
        resource_version=resource_version,
        entry=EntryDto(**entry.dict()),
    )



from typing import Dict


from karp.lex.domain import entities
from karp.lex.domain.entities.entry import EntryOp, EntryStatus
from karp.lex.domain.entities.resource import ResourceOp
from karp.lex.application.repositories import EntryUnitOfWork
from karp.db_infrastructure import db


class ResourceModel(db.Base):
    __tablename__ = "resources"
    history_id = db.Column(db.Integer, primary_key=True)
    entity_id = db.Column(db.ULIDType, nullable=False)
    resource_id = db.Column(db.String(32), nullable=False)
    resource_type = db.Column(db.String(32), nullable=False)
    version = db.Column(db.Integer, nullable=False)
    name = db.Column(db.String(64), nullable=False)
    entry_repo_id = db.Column(db.ULIDType, nullable=False)
    config = db.Column(db.NestedMutableJson, nullable=False)
    is_published = db.Column(db.Boolean, index=True, nullable=True, default=None)
    last_modified = db.Column(db.Float(precision=53), nullable=False)
    last_modified_by = db.Column(db.String(100), nullable=False)
    message = db.Column(db.String(100), nullable=False)
    op = db.Column(db.Enum(ResourceOp), nullable=False)
    # entry_json_schema = db.Column(db.Text, nullable=False)
    discarded = db.Column(db.Boolean, default=False)
    __table_args__ = (
        db.UniqueConstraint(
            "entity_id", "version", name="entity_id_version_unique_constraint"
        ),
        # mysql_character_set="utf8mb4",
        # TODO only one resource can be active, but several can be inactive
        #    here is how to do it in MariaDB, unclear whether this is possible using SQLAlchemy
        #    `virtual_column` char(0) as (if(active,'', NULL)) persistent
        #    and
        #    UNIQUE KEY `resource_version_unique_active` (`resource_id`,`virtual_column`)
        #    this works because the tuple (saldo, NULL) is not equal to (saldo, NULL)
    )

    def __repr__(self):
        return """<ResourceDTO(
                    history_id={},
                    entity_id={},
                    resource_id={},
                    version={},
                    name={},
                    config={},
                    entry_repo_id={},
                    is_published={},
                    last_modified={},
                    last_modified_by={},
                    discarded={},
                ) > """.format(
            self.history_id,
            self.entity_id,
            self.resource_id,
            self.version,
            self.name,
            self.config,
            self.entry_repo_id,
            self.is_published,
            self.last_modified,
            self.last_modified_by,
            self.discarded,
        )

    def to_entity(self) -> entities.Resource:
        return entities.Resource(
            entity_id=self.entity_id,
            resource_id=self.resource_id,
            version=self.version,
            name=self.name,
            config=self.config,
            entry_repo_id=self.entry_repo_id,
            is_published=self.is_published,
            last_modified=self.last_modified,
            last_modified_by=self.last_modified_by,
            discarded=self.discarded,
            message=self.message,
        )

    @staticmethod
    def from_entity(resource: entities.Resource) -> "ResourceModel":
        return ResourceModel(
            history_id=None,
            entity_id=resource.entity_id,
            resource_id=resource.resource_id,
            resource_type=resource.resource_type,
            version=resource.version,
            name=resource.name,
            config=resource.config,
            entry_repo_id=resource.entry_repository_id,
            is_published=resource.is_published,
            last_modified=resource.last_modified,
            last_modified_by=resource.last_modified_by,
            message=resource.message,
            op=resource.op,
            discarded=resource.discarded,
        )


class EntryUowModel(db.Base):
    __tablename__ = "entry_repos"
    history_id = db.Column(db.Integer, primary_key=True)
    entity_id = db.Column(db.ULIDType, nullable=False)
    type = db.Column(db.String(64), nullable=False)
    connection_str = db.Column(db.String(128))
    name = db.Column(db.String(64), nullable=False)
    config = db.Column(db.NestedMutableJson, nullable=False)
    last_modified = db.Column(db.Float(precision=53), nullable=False)
    last_modified_by = db.Column(db.String(100), nullable=False)
    message = db.Column(db.String(100), nullable=False)
    discarded = db.Column(db.Boolean, default=False)

    @staticmethod
    def from_entity(entry_uow: EntryUnitOfWork) -> "EntryUowModel":
        return EntryUowModel(
            history_id=None,
            entity_id=entry_uow.entity_id,
            type=entry_uow.repository_type,
            connection_str=entry_uow.connection_str,
            name=entry_uow.name,
            config=entry_uow.config,
            last_modified=entry_uow.last_modified,
            last_modified_by=entry_uow.last_modified_by,
            message=entry_uow.message,
            discarded=entry_uow.discarded,
        )


class BaseRuntimeEntry:
    entry_id = db.Column(
        # db.String(100, collation="utf8mb4_swedish_ci"), primary_key=True
        db.String(100),
        primary_key=True,
    )
    history_id = db.Column(db.Integer, nullable=False)
    entity_id = db.Column(db.ULIDType, nullable=False)
    discarded = db.Column(db.Boolean, nullable=False)


class BaseHistoryEntry:
    history_id = db.Column(db.Integer, primary_key=True)
    entity_id = db.Column(db.ULIDType, nullable=False)
    repo_id = db.Column(db.ULIDType, nullable=False)
    # entry_id = db.Column(db.String(100), nullable=False)
    version = db.Column(db.Integer, nullable=False)
    last_modified = db.Column(db.Float(53), nullable=False)
    last_modified_by = db.Column(db.String(100), nullable=False)
    body = db.Column(db.JSON, nullable=False)
    status = db.Column(db.Enum(EntryStatus), nullable=False)
    message = db.Column(db.Text(length=120))
    op = db.Column(db.Enum(EntryOp), nullable=False)
    discarded = db.Column(db.Boolean, default=False)

    @classmethod
    @db.declared_attr
    def __table_args__(cls):
        return db.UniqueConstraint(
            "entity_id", "version", name="id_version_unique_constraint"
        )

    @classmethod
    def from_entity(cls, entry: entities.Entry):
        return cls(
            history_id=None,
            entity_id=entry.entity_id,
            # entry_id=entry.entry_id,
            version=entry.version,
            last_modified=entry.last_modified,
            last_modified_by=entry.last_modified_by,
            body=entry.body,
            status=entry.status,
            message=entry.message,
            op=entry.op,
            discarded=entry.discarded,
            repo_id=entry.repo_id,
        )


# Dynamic models


def get_or_create_entry_history_model(resource_id: str) -> BaseHistoryEntry:
    history_table_name = create_history_table_name(resource_id)
    if history_table_name in class_cache:
        history_model = class_cache[history_table_name]
        # history_model.__table__.create(bind=db.engine, checkfirst=True)
        return history_model

    attributes = {
        "__tablename__": history_table_name,
        "__table_args__": None,  # (BaseHistoryEntry.__table_args__,),
        # "mysql_character_set": "utf8mb4",
    }

    sqlalchemy_class = type(history_table_name, (db.Base, BaseHistoryEntry), attributes)
    # sqlalchemy_class.__table__.create(bind=db.engine, checkfirst=True)
    class_cache[history_table_name] = sqlalchemy_class
    return sqlalchemy_class


def get_or_create_entry_runtime_model(
    resource_id: str, history_model: db.Table, config: Dict
) -> BaseRuntimeEntry:
    table_name = create_runtime_table_name(resource_id)

    if table_name in class_cache:
        runtime_model = class_cache[table_name]
        # runtime_model.__table__.create(bind=db.engine, checkfirst=True)
        # for child_model in runtime_model.child_tables.values():
        #     child_model.__table__.create(bind=db.engine, checkfirst=True)
        return runtime_model

    foreign_key_constraint = db.ForeignKeyConstraint(
        ["history_id"], [f"{history_model.__tablename__}.history_id"]
    )

    attributes = {
        "__tablename__": table_name,
        "__table_args__": (foreign_key_constraint,),
    }
    child_tables = {}

    for field_name in config.get("referenceable", ()):
        field = config["fields"][field_name]

        if not field.get("collection"):
            if field["type"] == "integer":
                column_type = db.Integer()
            elif field["type"] == "number":
                column_type = db.Float()
            elif field["type"] == "boolean":
                column_type = db.Boolean()
            elif field["type"] == "string":
                column_type = db.Text(32000)
            elif field["type"] == "long_string":
                column_type = db.Text(16000000)
            else:
                raise NotImplementedError()
            attributes[field_name] = db.Column(column_type)
        else:
            child_table_name = f"{table_name}_{field_name}"
            attributes[field_name] = db.relationship(
                child_table_name,
                backref=table_name,
                cascade="save-update,merge,delete,delete-orphan",
            )
            child_attributes = {
                "__tablename__": child_table_name,
                "__table_args__": (db.PrimaryKeyConstraint("entry_id", field_name),),
                "entry_id": db.Column(
                    db.String(100), db.ForeignKey(f"{table_name}.entry_id")
                ),
            }
            if field["type"] == "object":
                raise ValueError("not possible to reference lists of objects")
            if field["type"] == "integer":
                child_db_column_type = db.Integer()
            elif field["type"] == "number":
                child_db_column_type = db.Float()
            elif field["type"] == "string":
                child_db_column_type = db.String(100)
            else:
                raise NotImplementedError()
            child_attributes[field_name] = db.Column(child_db_column_type)
            child_class = type(child_table_name, (db.Base,), child_attributes)
            child_tables[field_name] = child_class

    sqlalchemy_class = type(
        table_name,
        (db.Base, BaseRuntimeEntry),
        attributes,
    )
    # sqlalchemy_class.__table__.create(bind=db.engine, checkfirst=True)
    sqlalchemy_class.child_tables = child_tables

    # for child_model in sqlalchemy_class.child_tables.values():
    #     child_model.__table__.create(bind=db.engine, checkfirst=True)
    class_cache[table_name] = sqlalchemy_class

    return sqlalchemy_class


class_cache = {}


# Helpers


def create_runtime_table_name(resource_id: str) -> str:
    return f"runtime_{resource_id}"


def create_history_table_name(resource_id: str) -> str:
    return resource_id



"""
Query API.

## Query DSL
"""
import logging
from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Path, Query, Security, status

from karp import auth, search
from karp.main import errors as karp_errors
from karp.search.application.queries import SearchService, QueryRequest
from karp.webapp import schemas

from karp.webapp import dependencies as deps
from karp.webapp.dependencies.fastapi_injector import inject_from_req


logger = logging.getLogger(__name__)


router = APIRouter()


@router.get(
    "/entries/{resource_id}/{entry_ids}",
    description="Returns a list of entries matching the given ids",
    name="Get lexical entries by id",
)
def get_entries_by_id(
    resource_id: str = Path(..., description="The resource to perform operation on"),
    entry_ids: str = Path(
        ...,
        description="Comma-separated. The ids to perform operation on.",
        regex=r"^\w(,\w)*",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    logger.debug("webapp.views.get_entries_by_id")
    if not auth_service.authorize(auth.PermissionLevel.read, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    return search_service.search_ids(resource_id, entry_ids)


@router.get("/split/{resources}", name="Query per resource")
def query_split(
    resources: str = Path(
        ...,
        regex=r"^[a-z_0-9\-]+(,[a-z_0-9\-]+)*$",
        description="A comma-separated list of resource identifiers",
    ),
    q: Optional[str] = Query(
        None,
        title="query",
        description="The query. If missing, all entries in chosen resource(s) will be returned.",
    ),
    from_: int = Query(
        0, alias="from", description="Specify which entry should be the first returned."
    ),
    size: int = Query(25, description="Number of entries in page."),
    sort: str = Query(
        None,
        description="The `field` to sort by. If missing, default order for each resource will be used.",
        regex=r"^[a-zA-Z0-9_\-]+(\|asc|desc)?",
    ),
    lexicon_stats: bool = Query(True, description="Show the hit count per lexicon"),
    include_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to return"
    ),
    exclude_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to remove from result"
    ),
    format: schemas.EntryFormat = Query(
        schemas.EntryFormat.json,
        description="Will return the result in the specified format.",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    logger.debug("/query/split called", extra={"resources": resources})
    resource_list = resources.split(",")
    if not auth_service.authorize(auth.PermissionLevel.read, user, resource_list):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    query_request = QueryRequest(
        resource_ids=resource_list,
        q=q,
        from_=from_,
        size=size,
        lexicon_stats=lexicon_stats,
    )
    try:
        response = search_service.query_split(query_request)
    except karp_errors.KarpError as err:
        logger.exception(
            "Error occured when calling '/query/split'",
            extra={"resources": resources, "q": q, "error_message": err.message},
        )
        raise
    except search.IncompleteQuery as err:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail={
                "errorCode": karp_errors.ClientErrorCodes.SEARCH_INCOMPLETE_QUERY,
                "message": "Error in query",
                "failing_query": err.failing_query,
                "error_description": err.error_description,
            },
        )
    return response


@router.get(
    "/{resources}",
    # summary="Returns a list of entries matching the given query in the given resources. The results are mixed from the given resources.",
    name="Query",
    responses={200: {"content": {"application/json": {}}}},
)
def query(
    resources: str = Path(
        ...,
        regex=r"^[a-z_0-9\-]+(,[a-z_0-9\-]+)*$",
        description="A comma-separated list of resource identifiers",
    ),
    q: Optional[str] = Query(
        None,
        title="query",
        description="The query. If missing, all entries in chosen resource(s) will be returned.",
    ),
    from_: int = Query(
        0, alias="from", description="Specify which entry should be the first returned."
    ),
    size: int = Query(25, description="Number of entries in page."),
    sort: List[str] = Query(
        [],
        description="The `field` to sort by. If missing, default order for each resource will be used.",
        regex=r"^[a-zA-Z0-9_\-]+(\|asc|desc)?",
    ),
    lexicon_stats: bool = Query(True, description="Show the hit count per lexicon"),
    include_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to return"
    ),
    exclude_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to remove from result"
    ),
    format_: schemas.EntryFormat = Query(
        schemas.EntryFormat.json,
        alias="format",
        description="Will return the result in the specified format.",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    """
    Returns a list of entries matching the given query in the given resources. The results are mixed from the given resources.

    """
    logger.debug(
        "Called 'query' called with",
        extra={"resources": resources, "from": from_, "size": size},
    )
    resource_list = resources.split(",")
    if not auth_service.authorize(auth.PermissionLevel.read, user, resource_list):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    query_request = QueryRequest(
        resource_ids=resource_list,
        q=q,
        from_=from_,
        size=size,
        sort=sort,
        include_fields=include_fields,
        exclude_fields=exclude_fields,
        format_=format_,
        lexicon_stats=lexicon_stats,
    )
    try:
        print(f"{search_service=}")
        response = search_service.query(query_request)

    except karp_errors.KarpError as err:
        logger.exception(
            "Error occured when calling 'query' with",
            extra={"resources": resources, "q": q, "error_message": err.message},
        )
        raise
    return response


def init_app(app):
    app.include_router(router)



"""Model for a lexical entry."""
import enum
import logging
import typing
from typing import Dict, List, Optional, Any

from deprecated import deprecated

from karp.foundation import constraints
from karp.lex.domain import errors, events
from karp.foundation.entity import TimestampedVersionedEntity
from karp.foundation.value_objects import unique_id

logger = logging.getLogger("karp")


class EntryOp(enum.Enum):
    ADDED = "ADDED"
    DELETED = "DELETED"
    UPDATED = "UPDATED"


class EntryStatus(enum.Enum):
    IN_PROGRESS = "IN-PROGRESS"
    IN_REVIEW = "IN_REVIEW"
    OK = "OK"


class Entry(TimestampedVersionedEntity):
    DiscardedEntityError = errors.DiscardedEntityError

    def __init__(
        self,
        *,
        # entry_id: str,
        body: Dict,
        message: str,
        # resource_id: str,
        repository_id: unique_id.UniqueId,
        # IN-PROGRESS, IN-REVIEW, OK, PUBLISHED
        status: EntryStatus = EntryStatus.IN_PROGRESS,
        op: EntryOp = EntryOp.ADDED,
        version: int = 1,
        **kwargs,
        # version: int = 0
    ):
        super().__init__(version=version, **kwargs)
        # self._entry_id = entry_id
        self._body = body
        self._op = op
        self._message = "Entry added." if message is None else message
        # self.resource_id = resource_id
        self._status = status
        self._repo_id = repository_id

    @property
    def repo_id(self) -> unique_id.UniqueId:
        return self._repo_id

    # @property
    # def entry_id(self):
    #     """The entry_id of this entry."""
    #     return self._entry_id

    # @entry_id.setter
    # @deprecated(version="6.0.7", reason="use update")
    # def entry_id(self, entry_id: str):
    #     self._check_not_discarded()
    #     self._entry_id = constraints.length_gt_zero("entry_id", entry_id)

    @property
    def body(self):
        """The body of the entry."""
        return self._body

    # @body.setter
    # @deprecated(version="6.0.7", reason="use update")
    def update_body(
        self,
        body: Dict,
        *,
        user: str,
        message: Optional[str] = None,
        timestamp: Optional[float] = None,
    ):
        self._check_not_discarded()
        self._body = self._update_field(body, user, timestamp)
        self._message = message or "Entry updated"
        self._op = EntryOp.UPDATED
        self._record_event(
            events.EntryUpdated(
                entity_id=self.id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                version=self.version,
                body=self.body,
                repo_id=self.repo_id,
                message=self.message,
            )
        )

    @property
    def op(self):
        """The latest operation of this entry."""
        return self._op

    @property
    def status(self):
        """The workflow status of this entry."""
        return self._status

    # @status.setter
    # @deprecated(version="6.0.7", reason="use update")
    # def status(self, status: EntryStatus):
    #     """The workflow status of this entry."""
    #     self._check_not_discarded()
    #     self._status = status

    @property
    def message(self):
        """The message for the latest operation of this entry."""
        return self._message

    def dict(self) -> Dict[str, Any]:
        return {
            # "entry_id": self._entry_id,
            "entity_id": self.entity_id,
            "resource": "",
            "version": self._version,
            "entry": self._body,
            "last_modified": self._last_modified,
            "last_modified_by": self._last_modified_by,
            "discarded": self._discarded,
            "message": self._message,
        }

    def discard(
        self,
        *,
        user: str,
        timestamp: Optional[float] = None,
        message: Optional[str] = None,
    ):
        if self._discarded:
            return
        self._op = EntryOp.DELETED
        self._message = message or "Entry deleted."
        self._discarded = self._update_field(True, user, timestamp)
        self._record_event(
            events.EntryDeleted(
                entity_id=self.id,
                # entry_id=self.entry_id,
                timestamp=self.last_modified,
                user=user,
                message=self._message,
                version=self.version,
                repo_id=self.repo_id,
            )
        )

    def _update_field(self, arg0, user: str, timestamp: Optional[float]):
        result = arg0
        self._last_modified_by = user
        self._last_modified = self._ensure_timestamp(timestamp)
        self._increment_version()
        return result

    def __repr__(self) -> str:
        return f"Entry(id={self._id}, version={self.version}, last_modified={self._last_modified}, body={self.body})"


# === Factories ===
def create_entry(
    # entry_id: str,
    body: Dict,
    *,
    entity_id: unique_id.UniqueId,
    repo_id: unique_id.UniqueId,
    last_modified_by: str = None,
    message: Optional[str] = None,
    last_modified: typing.Optional[float] = None,
) -> Entry:
    # if not isinstance(entry_id, str):
    #     entry_id = str(entry_id)
    entry = Entry(
        # entry_id=entry_id,
        body=body,
        message="Entry added." if not message else message,
        status=EntryStatus.IN_PROGRESS,
        op=EntryOp.ADDED,
        # entity_id=unique_id.make_unique_id(),
        version=1,
        last_modified_by="Unknown user" if not last_modified_by else last_modified_by,
        repository_id=repo_id,
        entity_id=entity_id,
        last_modified=last_modified,
    )
    entry.queue_event(
        events.EntryAdded(
            repo_id=repo_id,
            entity_id=entry.id,
            # entry_id=entry.entry_id,
            body=entry.body,
            message=entry.message or "",
            user=entry.last_modified_by,
            timestamp=entry.last_modified,
        )
    )
    return entry


# === Repository ===


# class EntryRepositorySettings:
#     """Settings for an EntryRepository."""
#
#     pass
#
#
# @singledispatch
# def create_entry_repository(settings: EntryRepositorySettings) -> EntryRepository:
#     raise RuntimeError(f"Don't know how to handle {settings!r}")



import pytest

from karp.foundation.value_objects import make_unique_id
from karp.foundation.events import EventBus, Event

from karp.lex_infrastructure import SqlResourceUnitOfWork
from karp.lex_infrastructure.repositories.sql_entries import SqlEntryUnitOfWork

from karp.tests.unit.lex import adapters, factories


class InMemoryEventBus(EventBus):
    def __init__(self) -> None:
        super().__init__()
        self.events: list[Event] = []

    def post(self, event: Event) -> None:
        self.events.append(event)


class TestSqlResourceUnitOfWork:
    def test_rolls_back_uncommitted_work_by_default(self, sqlite_session_factory):
        uow = SqlResourceUnitOfWork(
            session_factory=sqlite_session_factory, event_bus=InMemoryEventBus()
        )
        with uow:
            resource = factories.ResourceFactory()
            uow.resources.save(resource)

        new_session = sqlite_session_factory()
        rows = list(new_session.execute('SELECT * FROM "resources"'))
        assert rows == []

    def test_rolls_back_on_error(self, sqlite_session_factory):
        def do_something_that_fails(resource):
            print(resource.to_string())

        uow = SqlResourceUnitOfWork(
            session_factory=sqlite_session_factory, event_bus=InMemoryEventBus()
        )
        with pytest.raises(AttributeError):
            with uow:
                resource = factories.ResourceFactory()
                uow.resources.save(resource)
                do_something_that_fails(resource)

        new_session = sqlite_session_factory()
        rows = list(new_session.execute('SELECT * FROM "resources"'))
        assert rows == []


class MyException(Exception):
    pass


class TestSqlEntryUnitOfWork:
    def test_rolls_back_uncommitted_work_by_default(self, sqlite_session_factory):
        uow = SqlEntryUnitOfWork(
            # {"resource_id": "abc", "table_name": "abc"},
            # resource_config={"resource_id": "abc", "config": {}},
            sqlite_session_factory,
            event_bus=InMemoryEventBus(),
            name="test",
            config={},
            connection_str=":memory:",
            message="test message",
            entity_id=make_unique_id(),
        )
        with uow:
            entry = factories.EntryFactory()  # resource_id="abc")
            uow.entries.save(entry)

        new_session = sqlite_session_factory()
        rows = list(new_session.execute('SELECT * FROM "test"'))
        assert rows == []

    def test_rolls_back_on_error(self, sqlite_session_factory):

        uow = SqlEntryUnitOfWork(
            # {"resource_id": "abc", "table_name": "abc"},
            # resource_config={"resource_id": "abc", "config": {}},
            sqlite_session_factory,
            event_bus=InMemoryEventBus(),
            name="test",
            config={},
            connection_str=":memory:",
            message="test message",
            entity_id=make_unique_id(),
        )
        with pytest.raises(MyException):
            with uow:
                entry = factories.EntryFactory()
                uow.entries.save(entry)
                do_something_that_fails()

        new_session = sqlite_session_factory()
        rows = list(new_session.execute('SELECT * FROM "test"'))
        assert rows == []


def do_something_that_fails() -> None:
    raise MyException()



import logging
from typing import Optional
import sys

from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
import typer

from karp.main import bootstrap_app, modules, config
from karp.cliapp import subapps

logger = logging.getLogger(__name__)


def create_app():
    app = typer.Typer(help="Karp CLI")
    app_context = bootstrap_app()

    @app.callback()
    def set_app_context(
        ctx: typer.Context,
        version: Optional[bool] = typer.Option(
            None, "--version", callback=version_callback, is_eager=True
        ),
    ):
        if ctx.invoked_subcommand is None:
            ctx.obj = {}
        else:
            ctx.obj = {}
            ctx.obj["connection"] = app_context.container.get(Connection)
            ctx.obj["session"] = Session(bind=ctx.obj["connection"])
            logger.debug("create session", extra={"session": ctx.obj["session"]})
            ctx.obj["container"] = app_context.container.create_child_injector(
                modules.request_configuration(
                    conn=ctx.obj["connection"],
                    session=ctx.obj["session"],
                )
            )

    subapps.add_subapps(app)
    load_commands(app)

    return app


def version_callback(value: bool):
    if value:
        typer.echo(f"{config.PROJECT_NAME} CLI {config.VERSION}")
        raise typer.Exit()


def load_commands(app: typer.Typer):
    modules.load_modules("karp.clicommands", app=app)


cliapp = create_app()


if __name__ == "__main__":
    # cliapp = create_app()
    cliapp()



import pytest

from karp.lex.application import repositories
from karp.foundation.value_objects import unique_id
from . import adapters


def test_create_fake_entry_repo_uow(
    lex_ctx: adapters.UnitTestContext,
):
    uow_factory = lex_ctx.container.get(repositories.EntryRepositoryUnitOfWorkFactory)

    entry_repo_uow = uow_factory.create(
        "fake",
        unique_id.make_unique_id(),
        "name",
        {},
        connection_str=None,
        user="kristoff@example.com",
        message="msg",
        timestamp=123456789,
    )

    print(f"entry_repo_uow = {entry_repo_uow}")
    assert isinstance(entry_repo_uow, adapters.InMemoryEntryUnitOfWork)


@pytest.mark.skip()
def test_create_fake_entry_repo_uow2(
    lex_ctx: adapters.UnitTestContext,
):
    uow_factory = lex_ctx.container.get(repositories.EntryRepositoryUnitOfWorkFactory)

    entry_repo_uow = uow_factory.create("fake2", unique_id.make_unique_id(), "name", {})

    print(f"entry_repo_uow = {entry_repo_uow}")
    assert isinstance(entry_repo_uow, adapters.InMemoryEntryUnitOfWork2)



import abc
import logging
from typing import List, Generic, TypeVar, Any, Iterable

import injector
import logging


logger = logging.getLogger(__name__)


T = TypeVar("T")


class Event:
    pass


class EventMixin:
    def __init__(self) -> None:
        self._pending_domain_events: List[Event] = []

    def _record_event(self, event: Event) -> None:
        self._pending_domain_events.append(event)

    @property
    def domain_events(self) -> List[Event]:
        return self._pending_domain_events[:]

    def collect_new_events(self) -> Iterable[Event]:
        return self._pending_domain_events[:]

    def clear_events(self) -> None:
        self._pending_domain_events.clear()


class EventHandler(Generic[T]):
    """Simple generic used to associate handlers with events using DI.

    e.g EventHandler[ResourceCreated].
    """

    def __call__(self, event: T, *args: Any, **kwds: Any) -> Any:
        raise NotImplementedError()


class EventBus(abc.ABC):
    @abc.abstractmethod
    def post(self, event: Event) -> None:
        raise NotImplementedError


class InjectorEventBus(EventBus):
    def __init__(self, container: injector.Injector) -> None:
        self._container = container

    def post(self, event: Event) -> None:
        logger.info("handling event", extra={"karp_event": event})
        try:
            evt_handlers = self._container.get(
                List[EventHandler[type(event)]]  # type: ignore
            )
        except injector.UnsatisfiedRequirement as err:
            logger.info(
                "No event handler for event?",
                extra={"karp_event": event, "err_message": err},
            )
        else:
            for evt_handler in evt_handlers:
                logger.debug(
                    "handling event with handler",
                    extra={"karp_event": event, "evt_handler": evt_handler},
                )
                try:
                    evt_handler(event)
                except Exception as err:
                    logger.exception(
                        "Exception handling event", extra={"karp_event": event}
                    )
                    raise



from karp.foundation.value_objects.unique_id import UniqueId, make_unique_id
from karp.lex.domain.value_objects.entry_schema import EntrySchema
from karp.lex.domain.value_objects.resource_config import ResourceConfig

__all__ = ["UniqueId", "make_unique_id"]



import json
import logging
import re
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import elasticsearch
import elasticsearch.helpers  # pyre-ignore
import elasticsearch_dsl as es_dsl  # pyre-ignore
from tatsu import exceptions as tatsu_exc

from karp import search

# from karp import query_dsl
from karp.search.application.queries import (
    QueryRequest,
)
from karp.search.domain import errors
from karp.search.domain.errors import (
    UnsupportedField,
)  # IncompleteQuery,; UnsupportedQuery,
from karp.lex.domain.entities.entry import Entry
from karp.lex.domain.entities.resource import Resource
from karp.search.domain import query_dsl
from karp.search_infrastructure.elasticsearch6 import es_config
from karp.search_infrastructure.elasticsearch6 import Es6MappingRepository
from .es_query import EsQuery

# from karp.query_dsl import basic_ast as ast, op, is_a


logger = logging.getLogger(__name__)


class EsQueryBuilder(query_dsl.NodeWalker):
    def walk_object(self, node):
        return node

    def walk__and(self, node):
        result = self.walk(node.exps[0])
        for n in node.exps[1:]:
            result = result & self.walk(n)

        return result

    def walk__contains(self, node):
        return es_dsl.Q(
            "regexp", **{self.walk(node.field): f".*{self.walk(node.arg)}.*"}
        )

    def walk__endswith(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): f".*{self.walk(node.arg)}"})

    # def walk__equals(self, node):
    #     return es_dsl.Q(
    #         "match",
    #         **{
    #             self.walk(node.field): {"query": self.walk(node.arg), "operator": "and"}
    #         },
    #     )

    def walk__equals(self, node):
        return es_dsl.Q(
            "match",
            **{
                self.walk(node.field): {"query": self.walk(node.arg), "operator": "and"}
            },
        )

    def walk__exists(self, node):
        return es_dsl.Q("exists", field=self.walk(node.field))

    def walk__freergxp(self, node):
        return es_dsl.Q(
            "query_string", query=f"/{self.walk(node.arg)}/", default_field="*"
        )

    def walk__freetext_string(self, node):
        return es_dsl.Q("multi_match", query=self.walk(node.arg), fuzziness=1)

    def walk__freetext_any_but_string(self, node):
        return es_dsl.Q("multi_match", query=self.walk(node.arg))

    def walk_range(self, node):
        return es_dsl.Q(
            "range",
            **{self.walk(node.field): {self.walk(node.op): self.walk(node.arg)}},
        )

    walk__gt = walk_range
    walk__gte = walk_range
    walk__lt = walk_range
    walk__lte = walk_range

    def walk__missing(self, node):
        return es_dsl.Q(
            "bool", must_not=es_dsl.Q("exists", field=self.walk(node.field))
        )

    def walk__not(self, node):
        must_nots = [self.walk(expr) for expr in node.exps]
        return es_dsl.Q("bool", must_not=must_nots)

    def walk__or(self, node):
        result = self.walk(node.exps[0])
        for n in node.exps[1:]:
            result = result | self.walk(n)

        return result

    def walk__regexp(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): self.walk(node.arg)})

    def walk__startswith(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): f"{self.walk(node.arg)}.*"})


class Es6SearchService(search.SearchService):
    def __init__(
        self,
        es: elasticsearch.Elasticsearch,
        mapping_repo: Es6MappingRepository,
    ):
        self.es: elasticsearch.Elasticsearch = es
        self.mapping_repo = mapping_repo
        self.query_builder = EsQueryBuilder()
        self.parser = query_dsl.KarpQueryV6Parser(
            semantics=query_dsl.KarpQueryV6ModelBuilderSemantics()
        )

    def build_query(self, args, resource_str: str) -> EsQuery:
        query = EsQuery()
        query.parse_arguments(args, resource_str)
        return query

    def _format_result(self, resource_ids, response):
        logger.debug("_format_result called", extra={"resource_ids": resource_ids})
        resource_id_map = {
            resource_id: self.mapping_repo.get_name_base(resource_id)
            for resource_id in resource_ids
        }
        def format_entry(entry):

            dict_entry = entry.to_dict()
            version = dict_entry.pop("_entry_version", None)
            last_modified_by = dict_entry.pop("_last_modified_by", None)
            last_modified = dict_entry.pop("_last_modified", None)
            return {
                "id": entry.meta.id,
                "version": version,
                "last_modified": last_modified,
                "last_modified_by": last_modified_by,
                "resource": next(
                    resource
                    for resource, index_base in resource_id_map.items()
                    if entry.meta.index.startswith(index_base)
                ),
                "entry": dict_entry,
            }

        result = {
            "total": response.hits.total,
            "hits": [format_entry(entry) for entry in response],
        }
        return result

    def query(self, request: QueryRequest):
        logger.info("query called", extra={"request": request})
        query = EsQuery.from_query_request(request)
        return self.search_with_query(query)

    def query_split(self, request: QueryRequest):
        logger.info("query_split called", extra={"request": request})
        query = EsQuery.from_query_request(request)
        query.split_results = True
        return self.search_with_query(query)

    def search_with_query(self, query: EsQuery):
        logger.info("search_with_query called", extra={"query": query})
        es_query = None
        if query.q:
            try:
                model = self.parser.parse(query.q)
                es_query = self.query_builder.walk(model)
            except tatsu_exc.FailedParse as err:
                logger.debug("Parse error", extra={"err": err})
                raise errors.IncompleteQuery(
                    failing_query=query.q, error_description=str(err)
                ) from err
        if query.split_results:
            ms = es_dsl.MultiSearch(using=self.es)

            for resource in query.resources:
                alias_name = self.mapping_repo.get_alias_name(resource)
                s = es_dsl.Search(index=alias_name)

                if es_query is not None:
                    s = s.query(es_query)
                s = s[query.from_ : query.from_ + query.size]
                if query.sort:
                    s = s.sort(
                        *self.mapping_repo.translate_sort_fields([resource], query.sort)
                    )
                elif query.sort_dict and resource in query.sort_dict:
                    s = s.sort(
                        *self.translate_sort_fields(
                            [resource], query.sort_dict[resource]
                        )
                    )
                ms = ms.add(s)

            responses = ms.execute()
            result: dict[str, Any] = {"total": 0, "hits": {}}
            for i, response in enumerate(responses):
                result["hits"][query.resources[i]] = self._format_result(
                    query.resources, response
                ).get("hits", [])
                result["total"] += response.hits.total
                if query.lexicon_stats:
                    if "distribution" not in result:
                        result["distribution"] = {}
                    result["distribution"][query.resources[i]] = response.hits.total
        else:
            result = self._extracted_from_search_with_query_47(query, es_query)

        return result

    # TODO Rename this here and in `search_with_query`
    def _extracted_from_search_with_query_47(self, query, es_query):
        alias_names = [self.mapping_repo.get_alias_name(resource) for resource in query.resources]
        s = es_dsl.Search(using=self.es, index=alias_names, doc_type="entry")
        if es_query is not None:
            s = s.query(es_query)

        s = s[query.from_ : query.from_ + query.size]

        if query.lexicon_stats:
            s.aggs.bucket(
                "distribution", "terms", field="_index", size=len(query.resources)
            )
        if query.sort:
            s = s.sort(
                *self.mapping_repo.translate_sort_fields(query.resources, query.sort)
            )
        elif query.sort_dict:
            sort_fields = []
            for resource, sort in query.sort_dict.items():
                sort_fields.extend(
                    self.mapping_repo.translate_sort_fields([resource], sort)
                )
            s = s.sort(*sort_fields)
        logger.debug("s = %s", extra={"es_query s": s.to_dict()})
        response = s.execute()

        # TODO format response in a better way, because the whole response takes up too much space in the logs
        # logger.debug('response = {}'.format(response.to_dict()))

        logger.debug("calling _format_result")
        result = self._format_result(query.resources, response)
        if query.lexicon_stats:
            result["distribution"] = {}
            for bucket in response.aggregations.distribution.buckets:
                key = bucket["key"]
                result["distribution"][key.rsplit("_", 1)[0]] = bucket["doc_count"]

        return result

    def search_ids(self, resource_id: str, entry_ids: str):
        logger.info(
            "Called EsSearch.search_ids with:",
            extra={"resource_id": resource_id, "entry_ids": entry_ids},
        )
        entries = entry_ids.split(",")
        query = es_dsl.Q("terms", _id=entries)
        logger.debug("query", extra={"query": query})
        alias_name = self.mapping_repo.get_alias_name(resource_id)
        s = es_dsl.Search(using=self.es, index=alias_name).query(query)
        logger.debug("s", extra={"es_query s": s.to_dict()})
        response = s.execute()

        return self._format_result([resource_id], response)

    def statistics(self, resource_id: str, field: str) -> Iterable:
        alias_name = self.mapping_repo.get_alias_name(resource_id)
        s = es_dsl.Search(using=self.es, index=alias_name)
        s = s[:0]

        if field in self.mapping_repo.analyzed_fields[resource_id]:
            field += ".raw"

        logger.debug("Statistics: analyzed fields are:")
        logger.debug(json.dumps(self.mapping_repo.analyzed_fields, indent=4))
        logger.debug(
            "Doing aggregations on resource_id: {resource_id}, on field {field}".format(
                resource_id=resource_id, field=field
            )
        )
        s.aggs.bucket("field_values", "terms", field=field, size=2147483647)
        response = s.execute()
        logger.debug("Elasticsearch response", extra={"response": response})
        return [
            {"value": bucket["key"], "count": bucket["doc_count"]}
            for bucket in response.aggregations.field_values.buckets
        ]



from typing import Dict, Optional

from fastapi import APIRouter, Depends, HTTPException, Query, Security, status

from karp.auth import AuthService
from karp import auth, lex
from karp.foundation.value_objects import unique_id
from karp.foundation.value_objects.unique_id import UniqueIdStr

from karp.lex.application.queries import (
    EntryDiffDto,
    EntryDto,
    EntryDiffRequest,
    EntryHistoryRequest,
    GetEntryDiff,
    GetHistory,
    GetEntryHistory,
)
from karp.webapp import schemas

from karp.webapp import dependencies as deps
from karp.webapp.dependencies.fastapi_injector import inject_from_req


# pylint: disable=unsubscriptable-object

router = APIRouter()


@router.post("/diff/{resource_id}/{entry_id}", response_model=EntryDiffDto)
def get_diff(
    resource_id: str,
    entry_id: UniqueIdStr,
    user: auth.User = Security(deps.get_user, scopes=["admin"]),
    from_version: Optional[int] = None,
    to_version: Optional[int] = None,
    from_date: Optional[float] = None,
    to_date: Optional[float] = None,
    entry: Optional[Dict] = None,
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    get_entry_diff: GetEntryDiff = Depends(deps.get_entry_diff),
):
    if not auth_service.authorize(auth.PermissionLevel.admin, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="admin"'},
        )

    diff_request = EntryDiffRequest(
        resource_id=resource_id,
        entity_id=unique_id.parse(entry_id),
        from_version=from_version,
        to_version=to_version,
        from_date=from_date,
        to_date=to_date,
        entry=entry,
    )
    return get_entry_diff.query(diff_request)


@router.get(
    "/{resource_id}",
    response_model=lex.GetHistoryDto,
)
def get_history(
    resource_id: str,
    user: auth.User = Security(deps.get_user, scopes=["admin"]),
    user_id: Optional[str] = Query(None),
    entry_id: Optional[UniqueIdStr] = Query(None),
    from_date: Optional[float] = Query(None),
    to_date: Optional[float] = Query(None),
    to_version: Optional[int] = Query(None),
    from_version: Optional[int] = Query(None),
    current_page: int = Query(0),
    page_size: int = Query(100),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    get_history: GetHistory = Depends(deps.get_history),
):
    if not auth_service.authorize(auth.PermissionLevel.admin, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="admin"'},
        )
    history_request = EntryHistoryRequest(
        resource_id=resource_id,
        page_size=page_size,
        current_page=current_page,
        from_date=from_date,
        to_date=to_date,
        user_id=user_id,
        entry_id=entry_id,
        from_version=from_version,
        to_version=to_version,
    )
    return get_history.query(history_request)


def init_app(app):
    app.include_router(router)



import abc
import typing

from karp.foundation.value_objects import unique_id


class ResourceViews(abc.ABC):
    @abc.abstractmethod
    def get_resource_config(self, resource_id: str) -> typing.Dict:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_resource_ids(self, repo_id: unique_id.UniqueId) -> typing.List[str]:
        raise NotImplementedError()



from karp.auth.domain import errors
from karp.auth.domain.entities import User
from karp.auth.domain.auth_service import AuthService, AuthServiceConfig
from karp.auth.domain.value_objects import AccessToken
from karp.foundation.value_objects import PermissionLevel
from karp.auth.application.queries import (
    GetResourcePermissions,
    IsResourceProtected,
)

__all__ = [
    "errors",
    "User",
    "AuthService",
    "AuthServiceConfig",
    "AccessToken",
    "PermissionLevel",
    "GetResourcePermissions",
    "IsResourceProtected",
]



"""Unit tests for JWTAuthenticator"""
from datetime import timedelta
from typing import Dict, Optional
import os

import jwt

from karp.foundation.time import utc_now

from karp.auth import AccessToken
from karp.main.config import AUTH_JWT_AUDIENCE


PRIVATE_KEY_PATH = os.environ.get(
    "TEST_AUTH_JWT_PRIVATE_KEY_PATH", "karp/tests/data/private_key.pem"
)

with open(PRIVATE_KEY_PATH) as fp:
    jwt_private_key = fp.read()


def create_access_token(
    user: str,
    levels: Optional[Dict],
    scope: Optional[Dict] = None,
    priv_key: Optional[str] = None,
    audience: str = AUTH_JWT_AUDIENCE,
    expires_in: int = None,
) -> str:

    priv_key = priv_key or jwt_private_key
    expires_in = expires_in or 60

    iat = utc_now()
    time_delta = timedelta(minutes=abs(expires_in)).total_seconds()
    if expires_in < 0:
        exp = iat - time_delta
    else:
        exp = iat + time_delta
    token_payload = {
        "iss": "spraakbanken.gu.se",
        "iat": iat,
        "exp": exp,
    }
    if audience:
        token_payload["audience"] = audience
    if user:
        token_payload["sub"] = user
    if levels is not None:
        token_payload["levels"] = levels
    if scope:
        token_payload["scope"] = scope

    access_token = jwt.encode(token_payload, priv_key, algorithm="RS256")
    return access_token


def create_bearer_token(
    user: str,
    levels: Dict,
    scope: Optional[Dict] = None,
) -> AccessToken:
    return AccessToken(
        access_token=create_access_token(user, levels, scope),
        token_type="bearer",
    )



from karp import lex, search
from karp.foundation.value_objects.unique_id import make_unique_id
from karp.utility.time import utc_now


class GenericPreviewEntry(search.PreviewEntry):
    def __init__(
        self,
        resource_uow: lex.ResourceUnitOfWork,
        entry_transformer: search.EntryTransformer,
    ) -> None:
        super().__init__()
        self.entry_transformer = entry_transformer
        self.resource_uow = resource_uow

    def query(self, input_dto: search.PreviewEntryInputDto) -> search.EntryPreviewDto:
        with self.resource_uow:
            resource = self.resource_uow.repo.get_by_resource_id(input_dto.resource_id)

        # entry_id = resource.id_getter()(input_dto.entry)

        lex.EntrySchema(resource.entry_json_schema).validate_entry(input_dto.entry)
        input_entry = lex.EntryDto(
            # entry_id=entry_id,
            entity_id=make_unique_id(),
            resource=input_dto.resource_id,
            version=0,
            entry=input_dto.entry,
            last_modified=utc_now(),
            last_modified_by=input_dto.user,
        )
        entry = self.entry_transformer.transform(input_dto.resource_id, input_entry)
        return search.EntryPreviewDto(entry=entry)



import pytest
from starlette import status

from karp import auth
from karp.foundation.value_objects import unique_id
from karp.webapp.schemas import ResourceCreate, ResourcePublic


@pytest.fixture
def new_resource() -> ResourceCreate:
    return ResourceCreate(
        resource_id="test_resource",
        name="Test resource",
        message="test",
        config={
            "fields": {"foo": {"type": "string"}},
            "id": "foo",
        },
    )


class TestResourcesRoutes:
    def test_get_resource_exist(self, fa_data_client):
        response = fa_data_client.get("/resources/places")
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_get_resources_exist(self, fa_client):
        response = fa_client.get("/resources")
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_post_resources_exist(self, fa_client):
        response = fa_client.post("/resources")
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_get_resource_permissionss_exist(self, fa_client):
        response = fa_client.get("/resources/permissions")
        assert response.status_code != status.HTTP_404_NOT_FOUND


class TestGetResourcePermissions:
    def test_get_resources(self, fa_data_client):
        response = fa_data_client.get("/resources/permissions")

        assert response.status_code == status.HTTP_200_OK

        response_data = response.json()

        assert len(response_data) == 2
        assert response_data[0] == {
            "resource_id": "places",
            "protected": None,
        }
        assert response_data[1] == {
            "resource_id": "municipalities",
            "protected": "READ",
        }


class TestCreateResource:
    def test_missing_auth_header_returns_403(self, fa_client):
        response = fa_client.post("/resources/", json={})
        assert response.status_code == status.HTTP_403_FORBIDDEN

    def test_invalid_data_returns_422(self, fa_client, admin_token: auth.AccessToken):
        response = fa_client.post(
            "/resources/",
            json={},
            headers=admin_token.as_header(),
        )
        print(f"{response.json()=}")
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    def test_valid_input_creates_resource(
        self, fa_client, new_resource: ResourceCreate, admin_token: auth.AccessToken
    ):
        response = fa_client.post(
            "/resources/",
            json=new_resource.dict(),
            headers=admin_token.as_header(),
        )
        print(f"{response.json()=}")
        assert response.status_code == status.HTTP_201_CREATED

        created_resource = ResourceCreate(
            **response.json(),
        )
        expected_resource = new_resource
        expected_resource.version = 1
        expected_resource.is_published = False
        assert created_resource == expected_resource


class TestGetResource:
    def test_get_resource_by_resource_id(self, fa_client):
        response = fa_client.get("/resources/test_resource")
        print(f"{response.json()=}")
        assert response.status_code == status.HTTP_200_OK

        resource = ResourcePublic(**response.json())
        assert resource.resource_id == "test_resource"


class TestGetResources:
    def test_get_resources(self, fa_data_client):
        response = fa_data_client.get("/resources/")
        response_data = response.json()
        print(f"{response_data=}")
        assert response.status_code == status.HTTP_200_OK

        assert len(response_data) >= 3
        for resource_dict in response_data:
            ResourcePublic(**resource_dict)



from datetime import datetime
import logging
import re
from typing import Dict, Iterable, List, Optional, Any, Tuple, Union

import elasticsearch
from elasticsearch import exceptions as es_exceptions

from karp.foundation.events import EventBus

from karp.lex.domain.entities import Entry
from karp.search.application.repositories import (
    Index,
    IndexEntry,
    IndexUnitOfWork,
)
from karp.search.domain.errors import UnsupportedField
from karp.search_infrastructure.elasticsearch6 import Es6MappingRepository


logger = logging.getLogger(__name__)


class Es6Index(Index):
    def __init__(
        self,
        es: elasticsearch.Elasticsearch,
        mapping_repo: Es6MappingRepository,
    ):
        self.es = es
        self.mapping_repo = mapping_repo

    @property
    def seen(self):
        return []

    def create_index(self, resource_id: str, config):
        logger.info("creating es mapping")
        mapping = create_es6_mapping(config)

        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 1,
            "refresh_interval": -1,
        }
        if "settings" in mapping:
            settings |= mapping["settings"]
            del mapping["settings"]
        properties = mapping["properties"]
        properties["freetext"] = {"type": "text"}
        disabled_property = {"enabled": False}
        properties["_entry_version"] = disabled_property
        properties["_last_modified"] = disabled_property
        properties["_last_modified_by"] = disabled_property

        body = {
            "settings": settings,
            "mappings": {"entry": mapping},
        }

        index_alias_name = self.mapping_repo.create_index_and_alias_name(resource_id)
        logger.info("creating index", extra={"index_alias_name": index_alias_name, "body": body})
        result = self.es.indices.create(index=index_alias_name["index_name"], body=body)
        if "error" in result:
            logger.error(
                "failed to create index", extra={"index_name": index_alias_name["index_name"], "body": body}
            )
            raise RuntimeError("failed to create index")
        logger.info("index created")
        return index_alias_name

    def publish_index(self, resource_id: str):
        alias_name = self.mapping_repo.get_alias_name(resource_id)
        if self.es.indices.exists_alias(name=alias_name):
            self.es.indices.delete_alias(name=alias_name, index="*")

        index_name = self.mapping_repo.get_index_name(resource_id)
        self.mapping_repo.on_publish_resource(alias_name, index_name)
        logger.info(
            "publishing resource",
            extra={
                "resource_id": resource_id,
                "index_name": index_name,
                "alias_name": alias_name,
            },
        )
        self.es.indices.put_alias(name=alias_name, index=index_name)

    def add_entries(self, resource_id: str, entries: Iterable[IndexEntry]):
        index_name = self.mapping_repo.get_index_name(resource_id)
        index_to_es = []
        for entry in entries:
            assert isinstance(entry, IndexEntry)
            # entry.update(metadata.to_dict())
            index_to_es.append(
                {
                    "_index": index_name,
                    "_id": entry.id,
                    "_type": "entry",
                    "_source": entry.entry,
                }
            )

        elasticsearch.helpers.bulk(self.es, index_to_es, refresh=True)

    def delete_entry(
        self,
        resource_id: str,
        *,
        entry_id: Optional[str] = None,
        entry: Optional[Entry] = None,
    ):
        if not entry and not entry_id:
            raise ValueError("Must give either 'entry' or 'entry_id'.")
        if entry:
            entry_id = entry.entry_id
        logger.info(
            "deleting entry", extra={"entry_id": entry_id, "resource_id": resource_id}
        )
        index_name = self.mapping_repo.get_index_name(resource_id)
        try:
            self.es.delete(
                index=index_name,
                doc_type="entry",
                id=entry_id,
                refresh=True,
            )
        except elasticsearch.exceptions.ElasticsearchException:
            logger.exception(
                "Error deleting entry",
                extra={
                    "entry_id": entry_id,
                    "resource_id": resource_id,
                    "index_name": index_name,
                },
            )


def _create_es_mapping(config):
    es_mapping = {"dynamic": False, "properties": {}}

    fields = config["fields"]

    def recursive_field(parent_schema, parent_field_name, parent_field_def):
        if parent_field_def.get("virtual", False):
            fun = parent_field_def["function"]
            if list(fun.keys())[0] == "multi_ref":
                res_object = fun["multi_ref"]["result"]
                recursive_field(parent_schema, f"v_{parent_field_name}", res_object)
            if "result" in fun:
                res_object = fun["result"]
                recursive_field(parent_schema, f"v_{parent_field_name}", res_object)
            return
        if parent_field_def.get("ref"):
            if "field" in parent_field_def["ref"]:
                res_object = parent_field_def["ref"]["field"]
            else:
                res_object = {}
                res_object.update(parent_field_def)
                del res_object["ref"]
            recursive_field(parent_schema, f"v_{parent_field_name}", res_object)
        if parent_field_def["type"] != "object":
            # TODO this will not work when we have user defined types, s.a. saldoid
            # TODO number can be float/non-float, strings can be keyword or text in need of analyzing etc.
            if parent_field_def["type"] == "integer":
                mapped_type = "long"
            elif parent_field_def["type"] == "number":
                mapped_type = "double"
            elif parent_field_def["type"] == "boolean":
                mapped_type = "boolean"
            elif parent_field_def["type"] == "string":
                mapped_type = "text"
            elif parent_field_def["type"] == "long_string":
                mapped_type = "text"
            else:
                mapped_type = "keyword"
            result = {"type": mapped_type}
            if parent_field_def["type"] == "string":
                result["fields"] = {"raw": {"type": "keyword"}}
        else:
            result = {"properties": {}}

            for child_field_name, child_field_def in parent_field_def["fields"].items():
                recursive_field(result, child_field_name, child_field_def)

        parent_schema["properties"][parent_field_name] = result

    for field_name, field_def in fields.items():
        print(f"creating mapping for field '{field_name}'")
        recursive_field(es_mapping, field_name, field_def)

    return es_mapping


def create_es6_mapping(config: Dict) -> Dict:
    mapping = _create_es_mapping(config)
    mapping["settings"] = {
        "analysis": {
            "analyzer": {
                "default": {
                    "char_filter": [
                        "compound",
                        "swedish_aa",
                        "swedish_ae",
                        "swedish_oe",
                    ],
                    "filter": ["swedish_folding", "lowercase"],
                    "tokenizer": "standard",
                }
            },
            "char_filter": {
                "compound": {
                    "pattern": "-",
                    "replacement": "",
                    "type": "pattern_replace",
                },
                "swedish_aa": {
                    "pattern": "[Ǻǻ]",
                    "replacement": "å",
                    "type": "pattern_replace",
                },
                "swedish_ae": {
                    "pattern": "[æÆǞǟ]",
                    "replacement": "ä",
                    "type": "pattern_replace",
                },
                "swedish_oe": {
                    "pattern": "[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]",
                    "replacement": "ö",
                    "type": "pattern_replace",
                },
            },
            "filter": {
                "swedish_folding": {
                    "type": "icu_folding",
                    "unicodeSetFilter": "[^åäöÅÄÖ]",
                },
                "swedish_sort": {"language": "sv", "type": "icu_collation"},
            },
        }
    }
    return mapping


class Es6IndexUnitOfWork(IndexUnitOfWork):
    def __init__(
        self,
        es: elasticsearch.Elasticsearch,
        event_bus: EventBus,
        mapping_repo: Es6MappingRepository,
    ) -> None:
        super().__init__(event_bus=event_bus)
        self._index = Es6Index(
            es=es,
            mapping_repo=mapping_repo,
        )

    # @classmethod
    # def from_dict(cls, **kwargs):
    #     return cls()

    def _commit(self):
        logger.debug("Calling _commit in Es6IndexUnitOfWork")

    def rollback(self):
        return super().rollback()

    @property
    def repo(self) -> Es6Index:
        return self._index

    def _close(self):
        pass



import abc
from datetime import datetime
import logging
import re
from typing import Dict, List, Optional, Any, Tuple, Union

import elasticsearch
from elasticsearch import exceptions as es_exceptions

from karp.foundation.events import EventBus

from karp.lex.domain.entities import Entry
from karp.search.application.repositories import (
    Index,
    IndexEntry,
    IndexUnitOfWork,
)
from karp.search.domain.errors import UnsupportedField


logger = logging.getLogger("karp")

KARP_CONFIGINDEX = "karp_config"
KARP_CONFIGINDEX_TYPE = "configs"


class MappingRepository(abc.ABC):
    @abc.abstractmethod
    def get_index_name(self, resource_id: str) -> str:
        ...

    @abc.abstractmethod
    def get_alias_name(self, resource_id: str) -> str:
        ...


class Es6MappingRepository(MappingRepository):
    def __init__(
        self,
        es: elasticsearch.Elasticsearch,
        prefix: str = "",
    ):
        self.es = es
        self._prefix = prefix
        self._config_index = f"{prefix}_config" if prefix else KARP_CONFIGINDEX
        self.ensure_config_index_exist()
        analyzed_fields, sortable_fields = self._init_field_mapping()
        self.analyzed_fields: Dict[str, List[str]] = analyzed_fields
        self.sortable_fields: Dict[str, Dict[str, List[str]]] = sortable_fields

    def ensure_config_index_exist(self) -> None:
        if not self.es.indices.exists(index=self._config_index):
            self.es.indices.create(
                index=self._config_index,
                body={
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 1,
                        "refresh_interval": -1,
                    },
                    "mappings": {
                        KARP_CONFIGINDEX_TYPE: {
                            "dynamic": False,
                            "properties": {
                                "index_name": {"type": "text"},
                                "alias_name": {"type": "text"},
                            },
                        }
                    },
                },
            )

    def create_index_name(self, resource_id: str) -> str:
        date = datetime.now().strftime("%Y-%m-%d-%H%M%S%f")
        return f"{self._prefix}{resource_id}_{date}"

    def create_alias_name(self, resource_id: str) -> str:
        return f"{self._prefix}{resource_id}"

    def create_index_and_alias_name(self, resource_id: str) -> dict[str, str]:
        return self._update_config(resource_id)

    def get_name_base(self, resource_id: str) -> str:
        return f"{self._prefix}{resource_id}"

    def _update_config(self, resource_id: str) -> dict[str, str]:
        index_name = self.create_index_name(resource_id)
        alias_name = self.create_alias_name(resource_id)
        names = {"index_name": index_name, "alias_name": alias_name}
        self.es.index(
            index=self._config_index,
            id=resource_id,
            doc_type=KARP_CONFIGINDEX_TYPE,
            body=names,
        )
        return names

    def get_index_name(self, resource_id: str) -> str:
        try:
            res = self.es.get(
                index=self._config_index, id=resource_id, doc_type=KARP_CONFIGINDEX_TYPE
            )
        except es_exceptions.NotFoundError as err:
            logger.info(
                "didn't find index_name for resource '%s' details: %s", resource_id, err
            )
            return self._update_config(resource_id)["index_name"]
        return res["_source"]["index_name"]

    def get_alias_name(self, resource_id: str) -> str:
        try:
            res = self.es.get(
                index=self._config_index, id=resource_id, doc_type=KARP_CONFIGINDEX_TYPE
            )
        except es_exceptions.NotFoundError as err:
            logger.info(
                "didn't find alias_name for resource '%s' details: %s", resource_id, err
            )
            return self._update_config(resource_id)["alias_name"]
        return res["_source"]["alias_name"]

    @staticmethod
    def get_analyzed_fields_from_mapping(
        properties: Dict[str, Dict[str, Dict[str, Any]]]
    ) -> List[str]:
        analyzed_fields = []

        for prop_name, prop_values in properties.items():
            if "properties" in prop_values:
                res = Es6MappingRepository.get_analyzed_fields_from_mapping(
                    prop_values["properties"]
                )
                analyzed_fields.extend([f"{prop_name}.{prop}" for prop in res])
            elif prop_values["type"] == "text":
                analyzed_fields.append(prop_name)
        return analyzed_fields

    def _init_field_mapping(
        self,
    ) -> Tuple[Dict[str, List[str]], Dict[str, Dict[str, List[str]]]]:
        """
        Create a field mapping based on the mappings of elasticsearch
        currently the only information we need is if a field is analyzed (i.e. text)
        or not.
        """

        field_mapping: Dict[str, List[str]] = {}
        sortable_fields = {}
        # Doesn't work for tests, can't find resource_definition
        # for resource in resourcemgr.get_available_resources():
        #     mapping = self.es.indices.get_mapping(index=resource.resource_id)
        #     field_mapping[resource.resource_id] = parse_mapping(
        #         next(iter(mapping.values()))['mappings']['entry']['properties']
        #     )
        aliases = self._get_all_aliases()
        mapping: Dict[
            str, Dict[str, Dict[str, Dict[str, Dict]]]
        ] = self.es.indices.get_mapping()
        # print(f"mapping = {mapping}")
        for (alias, index) in aliases:
            if (
                "mappings" in mapping[index]
                and "entry" in mapping[index]["mappings"]
                and "properties" in mapping[index]["mappings"]["entry"]
            ):
                field_mapping[
                    alias
                ] = Es6MappingRepository.get_analyzed_fields_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
                sortable_fields[
                    alias
                ] = Es6MappingRepository.create_sortable_map_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
        return field_mapping, sortable_fields

    def _get_index_mappings(
        self, index: Optional[str] = None
    ) -> Dict[str, Dict[str, Dict[str, Dict[str, Dict]]]]:
        kwargs = {"index": index} if index is not None else {}
        return self.es.indices.get_mapping(**kwargs)

    def _get_all_aliases(self) -> List[Tuple[str, str]]:
        """
        :return: a list of tuples (alias_name, index_name)
        """
        result = self.es.cat.aliases(h="alias,index")
        logger.debug(f"{result}")
        index_names = []
        for index_name in result.split("\n")[:-1]:
            logger.debug(f"index_name = {index_name}")
            if index_name[0] != ".":
                if match := re.search(r"([^ ]*) +(.*)", index_name):
                    groups = match.groups()
                    alias = groups[0]
                    index = groups[1]
                    index_names.append((alias, index))
        return index_names

    def translate_sort_fields(
        self, resources: List[str], sort_values: List[str]
    ) -> List[Union[str, Dict[str, Dict[str, str]]]]:
        """Translate sort field to ES sort fields.

        Arguments:
            sort_values {List[str]} -- values to sort by

        Returns:
            List[str] -- values that ES can sort by.
        """
        translated_sort_fields: List[Union[str, Dict[str, Dict[str, str]]]] = []
        for sort_value in sort_values:
            sort_order = None
            if "|" in sort_value:
                sort_value, sort_order = sort_value.split("|", 1)
            for resource_id in resources:
                if sort_order:
                    translated_sort_fields.extend(
                        (
                            {field: {"order": sort_order}}
                            for field in self.translate_sort_field(
                                resource_id, sort_value
                            )
                        )
                    )
                translated_sort_fields.extend(
                    self.translate_sort_field(resource_id, sort_value)
                )

        return translated_sort_fields

    def translate_sort_field(self, resource_id: str, sort_value: str) -> List[str]:
        logger.debug(
            f"es6_indextranslate_sort_field: sortable_fields[{resource_id}] = {self.sortable_fields[resource_id]}"
        )
        if sort_value in self.sortable_fields[resource_id]:
            return self.sortable_fields[resource_id][sort_value]
        else:
            raise UnsupportedField(
                f"You can't sort by field '{sort_value}' for resource '{resource_id}'"
            )

    def on_publish_resource(self, alias_name: str, index_name: str):
        mapping = self._get_index_mappings(index=index_name)
        if (
            "mappings" in mapping[index_name]
            and "entry" in mapping[index_name]["mappings"]
            and "properties" in mapping[index_name]["mappings"]["entry"]
        ):
            self.analyzed_fields[
                alias_name
            ] = Es6MappingRepository.get_analyzed_fields_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )
            self.sortable_fields[
                alias_name
            ] = Es6MappingRepository.create_sortable_map_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )

    @staticmethod
    def create_sortable_map_from_mapping(properties: Dict) -> Dict[str, List[str]]:
        sortable_map = {}

        def parse_prop_value(sort_map, base_name, prop_name, prop_value: Dict):
            if "properties" in prop_value:
                for ext_name, ext_value in prop_value["properties"].items():
                    ext_base_name = f"{base_name}.{ext_name}"
                    parse_prop_value(sort_map, ext_base_name, ext_base_name, ext_value)
                return
            if prop_value["type"] in [
                "boolean",
                "date",
                "double",
                "keyword",
                "long",
                "ip",
            ]:
                sort_map[base_name] = [prop_name]
                sort_map[prop_name] = [prop_name]
                return
            if prop_value["type"] == "text":
                if "fields" in prop_value:
                    for ext_name, ext_value in prop_value["fields"].items():
                        parse_prop_value(
                            sort_map, base_name, f"{base_name}.{ext_name}", ext_value
                        )
                return

        for prop_name, prop_value in properties.items():
            parse_prop_value(sortable_map, prop_name, prop_name, prop_value)
            # if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
            #     sortable_map[prop_name] = prop_name
            # if prop_value["type"] == "text":
            #     if "fields" in prop_value:

        return sortable_map



import injector
import pytest

from karp.foundation.commands import CommandBus
from karp.lex import Lex
from karp.lex.domain import commands as lex_commands
from karp.lex.application.repositories import (
    EntryUowRepositoryUnitOfWork,
    EntryRepositoryUnitOfWorkFactory,
)
from karp.main.modules import CommandBusMod, EventBusMod
from karp.lex.application import repositories

from .adapters import (
    InMemoryEntryUowRepositoryUnitOfWork,
    # InMemoryEntryUowFactory,
    InMemoryResourceUnitOfWork,
)
from . import adapters, factories


@pytest.fixture(name="entry_repo_repo_uow")
def fixture_entry_repo_repo_uow() -> EntryUowRepositoryUnitOfWork:
    return InMemoryEntryUowRepositoryUnitOfWork()


@pytest.fixture(name="create_entry_repository")
def fixture_create_entry_repository() -> lex_commands.CreateEntryRepository:
    return factories.CreateEntryRepositoryFactory()


@pytest.fixture(name="resource_uow")
def fixture_resource_uow() -> InMemoryResourceUnitOfWork:
    return InMemoryResourceUnitOfWork()


# @pytest.fixture(name="entry_uows")
# def fixture_entry_uows() -> repositories.EntriesUnitOfWork:
#     return repositories.EntriesUnitOfWork()


# @pytest.fixture(name="entry_uow_factory")
# def fixture_entry_uow_factory() -> InMemoryEntryUowFactory:
#     return InMemoryEntryUowFactory()


@pytest.fixture()
def lex_ctx() -> adapters.UnitTestContext:
    container = injector.Injector(
        [
            CommandBusMod(),
            EventBusMod(),
            Lex(),
            adapters.InMemoryLexInfrastructure(),
        ],
        auto_bind=False,
    )
    return adapters.UnitTestContext(
        container=container,
        command_bus=container.get(CommandBus),  # type: ignore [misc]
    )


@pytest.fixture()
def command_bus(unit_test_injector: injector.Injector) -> CommandBus:
    return unit_test_injector.get(CommandBus)  # type: ignore [misc]



from karp.lex.domain.entities.entry import create_entry
from karp.foundation.value_objects import unique_id
from karp.lex_infrastructure.queries import generic_network

from karp.tests.unit.lex import factories


def test__create_ref():
    resource_id = "resource_id"
    resource_version = 1
    _id = 5
    entry_id = "entry_id"
    entry_body = {"body": {"of": "entry"}}
    entry = factories.EntryFactory(body=entry_body)

    ref = generic_network._create_ref(resource_id, resource_version, entry)

    assert ref.resource_id == resource_id
    assert ref.resource_version == resource_version
    # assert ref["entry"]["id"] == _id
    assert ref.entry.entity_id == entry.id
    assert ref.entry.entry == entry_body



import logging
from typing import Dict, Optional, Iterable

from karp.search.domain import query_dsl
from karp.search.application.repositories import Index, IndexUnitOfWork, IndexEntry
from karp.lex.domain.entities.entry import Entry
from karp.lex.domain.entities.resource import Resource

logger = logging.getLogger(__name__)


class SqlSearchService(Index):
    def __init__(self):
        self.seen = []
        self.parser = query_dsl.KarpQueryV6Parser(
            semantics=query_dsl.KarpQueryV6ModelBuilderSemantics()
        )

    def create_index(self, resource_id: str, resource_config: Dict):
        pass

    def publish_index(self, resource_id: str):
        pass

    def query(self):
        pass

    def query_split(self):
        pass

    def search_ids(self):
        pass

    def statistics(self):
        pass

    def add_entries(self, resource_id, entries: Iterable[IndexEntry]):
        pass

    def delete_entry(
        self,
        resource: Resource,
        *,
        entry: Optional[Entry] = None,
        entry_id: Optional[str] = None,
    ):
        pass


class SqlIndexUnitOfWork(IndexUnitOfWork):
    @classmethod
    def from_dict(cls, **kwargs):
        logger.debug(f"SqlIndexUnitOfWork.from_dict: kwargs = {kwargs}")
        return cls(**kwargs)

    def __init__(
        self,
        session_factory,
        event_bus,
    ):
        super().__init__(event_bus)
        # session_factory=DEFAULT_SESSION_FACTORY):
        self.session_factory = session_factory
        self._index = SqlSearchService()

    def _commit(self):
        pass

    def rollback(self):
        pass

    @property
    def repo(self):
        if not self._index:
            raise RuntimeError()
        return self._index



import typing

import pydantic

from karp.foundation.value_objects import unique_id
from .base import Command
from karp.utility import time

from karp.lex.domain import errors
from karp.lex.domain.value_objects import UniqueId


class AddEntry(Command):
    entity_id: unique_id.UniqueId = pydantic.Field(
        default_factory=unique_id.make_unique_id
    )
    resource_id: str
    entry: typing.Dict
    user: str
    message: str


class AddEntries(Command):
    resource_id: str
    entries: typing.Iterable[typing.Dict]
    user: str
    message: str
    chunk_size: int = 0


class AddEntriesInChunks(AddEntries):
    chunk_size: int


class DeleteEntry(Command):
    resource_id: str
    entity_id: UniqueId
    user: str
    message: typing.Optional[str] = None


class ImportEntries(AddEntries):
    pass


class ImportEntriesInChunks(AddEntriesInChunks):
    pass


class UpdateEntry(Command):
    resource_id: str
    entity_id: UniqueId
    # entry_id: str
    version: int
    entry: typing.Dict
    user: str
    message: str
    resource_version: typing.Optional[int] = None
    force: bool = False



class DomainError(Exception):
    pass


class UnsupportedField(DomainError):
    pass


class UnsupportedQuery(DomainError):
    pass


class IncompleteQuery(DomainError):
    def __init__(
        self, failing_query: str, error_description: str, *args: object
    ) -> None:
        super().__init__(*args)
        self.failing_query = failing_query
        self.error_description = error_description



import logging

import fastjsonschema

from karp.lex.domain import errors


logger = logging.getLogger(__name__)


class EntrySchema:
    def __init__(self, json_schema: dict):
        try:
            self._compiled_schema = fastjsonschema.compile(json_schema)
        except fastjsonschema.JsonSchemaDefinitionException as e:
            raise errors.InvalidEntrySchema() from e

    def validate_entry(self, json_obj: dict):
        try:
            self._compiled_schema(json_obj)
        except fastjsonschema.JsonSchemaException as e:
            logger.warning(
                "Entry not valid", extra={"entry": json_obj, "error_message": str(e)}
            )
            raise errors.InvalidEntry() from e



import dataclasses
from typing import Dict, Iterable, List, Optional

import injector

from karp.foundation.commands import CommandBus
from karp.foundation.events import EventBus
from karp.foundation.time import utc_now
from karp.lex.application.repositories import entries
from karp.lex.domain.entities import entry
from karp.search.application.repositories import IndexUnitOfWork, Index, IndexEntry
from karp.tests.foundation.adapters import InMemoryUnitOfWork


@dataclasses.dataclass
class SearchUnitTestContext:
    container: injector.Injector
    command_bus: CommandBus
    event_bus: EventBus


class InMemoryIndex(Index):
    @dataclasses.dataclass
    class Index:
        config: Dict
        created_at: float
        entries: Dict[str, IndexEntry] = dataclasses.field(default_factory=dict)
        created: bool = True
        published: bool = False

    def __init__(self) -> None:
        super().__init__()
        self.indicies: dict[str, InMemoryIndex.Index] = {}
        self.seen = []

    def create_index(self, resource_id: str, config: Dict):
        self.indicies[resource_id] = InMemoryIndex.Index(
            config=config, created_at=utc_now()
        )

    def publish_index(self, alias_name: str, index_name: str = None):
        self.indicies[alias_name].published = True

    def add_entries(self, resource_id: str, entries: Iterable[IndexEntry]):
        for entry in entries:
            self.indicies[resource_id].entries[entry.id] = entry

    def delete_entry(
        self,
        resource_id: str,
        *,
        entry_id: Optional[str] = None,
        # entry: typing.Optional[model.Entry]
    ):
        if entry_id is None:
            return
        if not isinstance(entry_id, str):
            entry_id = str(entry_id)
        del self.indicies[resource_id].entries[entry_id]

    #    def search_ids(self, resource_id: str, entry_ids: str):
    #        return {}
    #
    #    def query(self, request: search_service.QueryRequest):
    #        return {}
    #
    #    def query_split(self, request: search_service.QueryRequest):
    #        return {}
    #
    #    def statistics(self, resource_id: str, field: str):
    #        return {}
    def num_entities(self) -> int:
        return sum(len(entries) for entries in self.indicies.values())


class InMemoryIndexUnitOfWork(InMemoryUnitOfWork, IndexUnitOfWork):
    def __init__(self):
        # super().__init__()
        self._index = InMemoryIndex()

    @property
    def repo(self) -> InMemoryIndex:
        return self._index


class InMemorySearchInfrastructure(injector.Module):
    @injector.provider
    @injector.singleton
    def index_uow(self) -> IndexUnitOfWork:
        return InMemoryIndexUnitOfWork()



from karp.auth.application.queries import IsResourceProtected
from karp.auth import PermissionLevel


class InMemoryIsResourceProtected(IsResourceProtected):
    def query(self, resource_id: str, level: PermissionLevel) -> bool:
        return super().query(resource_id, level)



import typing

from sb_json_tools import jsondiff
import logging

from karp import lex
from karp.lex import GetHistoryDto, HistoryDto
from karp.lex.domain import errors

from karp.lex.domain.entities import Entry
from karp.lex.application.queries import (
    EntryViews,
    EntryDto,
    EntryDiffDto,
    GetEntryDiff,
    GetEntryHistory,
    GetHistory,
    EntryHistoryRequest,
    EntryDiffRequest,
    GetEntryRepositoryId,
)
from karp.foundation.value_objects import unique_id
from karp.lex.application.repositories import EntryUowRepositoryUnitOfWork


logger = logging.getLogger(__name__)


class GenericEntryViews(EntryViews):
    def __init__(
        self,
        get_entry_repo_id: GetEntryRepositoryId,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> None:
        super().__init__()
        self.get_entry_repo_id = get_entry_repo_id
        self.entry_repo_uow = entry_repo_uow

    def get_by_id(self, resource_id: str, entity_id: unique_id.UniqueId) -> EntryDto:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return self._entry_to_entry_dto(uw.repo.by_id(entity_id), resource_id)

    def get_by_id_optional(
        self, resource_id: str, entity_id: unique_id.UniqueId
    ) -> typing.Optional[EntryDto]:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            if entry := uw.repo.get_by_id_optional(entity_id):
                return self._entry_to_entry_dto(entry, resource_id)
        return None

    # def get_by_entry_id(self, resource_id: str, entry_id: str) -> EntryDto:
    #     entry_repo_id = self.get_entry_repo_id.query(resource_id)
    #     with self.entry_repo_uow as uw:
    #         entry_uow = uw.repo.get_by_id(entry_repo_id)
    #     with entry_uow as uw:
    #         return self._entry_to_entry_dto(uw.repo.by_entry_id(entry_id), resource_id)

    # def get_by_entry_id_optional(
    #     self, resource_id: str, entry_id: str
    # ) -> typing.Optional[EntryDto]:
    #     entry_repo_id = self.get_entry_repo_id.query(resource_id)
    #     with self.entry_repo_uow as uw:
    #         entry_uow = uw.repo.get_by_id(entry_repo_id)
    #     with entry_uow as uw:
    #         entry = uw.repo.get_by_entry_id_optional(entry_id)
    #         if entry:
    #             return self._entry_to_entry_dto(entry, resource_id)
    #     return None

    def get_total(self, resource_id: str) -> int:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return uw.repo.num_entities()

    def get_by_referenceable(self, resource_id: str, filters):
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return (
                self._entry_to_entry_dto(entry, resource_id)
                for entry in uw.repo.by_referenceable(filters)
            )

    def all_entries(self, resource_id: str) -> typing.Iterable[EntryDto]:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return (
                self._entry_to_entry_dto(entry, resource_id)
                for entry in uw.repo.all_entries()
            )

    def _entry_to_entry_dto(self, entry: Entry, resource_id: str) -> EntryDto:
        return EntryDto(
            # entry_id=entry.entry_id,
            entity_id=entry.entity_id,
            resource=resource_id,
            version=entry.version,
            entry=entry.body,
            last_modified=entry.last_modified,
            last_modified_by=entry.last_modified_by,
        )


class GenericEntryQuery:
    def __init__(
        self,
        resource_uow: lex.ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ):
        self._resource_uow = resource_uow
        self.entry_repo_uow = entry_repo_uow

    def get_entry_repo_id(self, resource_id: str) -> unique_id.UniqueId:
        with self._resource_uow as uw:
            return uw.repo.by_resource_id(resource_id).entry_repository_id


class GenericGetEntryHistory(GenericEntryQuery, GetEntryHistory):
    def query(
        self,
        resource_id: str,
        entity_id: unique_id.UniqueId,
        # entry_id: str,
        version: typing.Optional[int],
    ) -> EntryDto:
        entry_repo_id = self.get_entry_repo_id(resource_id)
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            entry_repo_id
        ) as uw:
            result = uw.repo.by_id(entity_id, version=version)

        return EntryDto(
            # entry_id=entry_id,
            entity_id=result.id,
            resource=resource_id,
            version=result.version,
            entry=result.body,
            last_modified_by=result.last_modified_by,
            last_modified=result.last_modified,
        )


class GenericGetHistory(GenericEntryQuery, GetHistory):
    def query(
        self,
        request: EntryHistoryRequest,
    ) -> GetHistoryDto:
        logger.info("querying history", extra={"request": request})
        entry_repo_id = self.get_entry_repo_id(request.resource_id)
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            entry_repo_id
        ) as uw:
            paged_query, total = uw.repo.get_history(
                entry_id=request.entry_id,
                user_id=request.user_id,
                from_date=request.from_date,
                to_date=request.to_date,
                from_version=request.from_version,
                to_version=request.to_version,
                offset=request.current_page * request.page_size,
                limit=request.page_size,
            )
        result = []
        previous_body: dict[str, typing.Any] = {}
        for history_entry in paged_query:
            # TODO fix this, we should get the diff in another way, probably store the diffs directly in the database
            # entry_version = history_entry.version
            # if entry_version > 1:
            #     previous_entry = uw.repo.by_entry_id(
            #         history_entry.entry_id, version=entry_version - 1
            #     )
            #     previous_body = previous_entry.body
            # else:
            #     previous_body = {}
            history_diff = jsondiff.compare(previous_body, history_entry.body)
            logger.info("diff", extra={"diff": history_diff})
            result.append(
                HistoryDto(
                    timestamp=history_entry.last_modified,
                    message=history_entry.message or "",
                    entity_id=history_entry.entity_id,
                    # entry_id=history_entry.entry_id,
                    version=history_entry.version,
                    op=history_entry.op,
                    user_id=history_entry.last_modified_by,
                    diff=history_diff,
                )
            )
            previous_body = history_entry.body

        return GetHistoryDto(history=result, total=total)


class GenericGetEntryDiff(GenericEntryQuery, GetEntryDiff):
    def query(
        self,
        request: EntryDiffRequest,
    ) -> EntryDiffDto:

        entry_repo_id = self.get_entry_repo_id(request.resource_id)
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            entry_repo_id
        ) as uw:
            db_entry = uw.repo.by_id(request.entity_id)

            #     src = resource_obj.model.query.filter_by(entry_id=entry_id).first()
            #
            #     query = resource_obj.history_model.query.filter_by(entry_id=src.id)
            #     timestamp_field = resource_obj.history_model.timestamp
            #
            if request.from_version:
                obj1 = uw.repo.by_id(db_entry.id, version=request.from_version)
            elif request.from_date is not None:
                obj1 = uw.repo.by_id(db_entry.id, after_date=request.from_date)
            else:
                obj1 = uw.repo.by_id(db_entry.id, oldest_first=True)

            obj1_body = obj1.body if obj1 else None

            if request.to_version:
                obj2 = uw.repo.by_id(db_entry.id, version=request.to_version)
                obj2_body = obj2.body
            elif request.to_date is not None:
                obj2 = uw.repo.by_id(db_entry.id, before_date=request.to_date)
                obj2_body = obj2.body
            elif request.entry is not None:
                obj2 = None
                obj2_body = request.entry
            else:
                obj2 = db_entry
                obj2_body = db_entry.body

        #     elif from_date is not None:
        #         obj1_query = query.filter(timestamp_field >= from_date).order_by(
        #             timestamp_field
        #         )
        #     else:
        #         obj1_query = query.order_by(timestamp_field)
        #     if to_version:
        #         obj2_query = query.filter_by(version=to_version)
        #     elif to_date is not None:
        #         obj2_query = query.filter(timestamp_field <= to_date).order_by(
        #             timestamp_field.desc()
        #         )
        #     else:
        #         obj2_query = None
        #
        #     obj1 = obj1_query.first()
        #     obj1_body = json.loads(obj1.body) if obj1 else None
        #
        #     if obj2_query:
        #         obj2 = obj2_query.first()
        #         obj2_body = json.loads(obj2.body) if obj2 else None
        #     elif entry is not None:
        #         obj2 = None
        #         obj2_body = entry
        #     else:
        #         obj2 = query.order_by(timestamp_field.desc()).first()
        #         obj2_body = json.loads(obj2.body) if obj2 else None
        #
        if not obj1_body or not obj2_body:
            raise errors.DiffImposible("diff impossible!")
        #
        return EntryDiffDto(
            diff=jsondiff.compare(obj1_body, obj2_body),
            from_version=obj1.version,
            to_version=obj2.version if obj2 else None,
        )



import typer

from karp.cliapp.subapps import (
    database_subapp,
    entries_subapp,
    entry_repo_subapp,
    query_subapp,
    resource_subapp,
)


def add_subapps(app: typer.Typer) -> None:
    database_subapp.init_app(app)
    entries_subapp.init_app(app)
    entry_repo_subapp.init_app(app)
    query_subapp.init_app(app)
    resource_subapp.init_app(app)



import dataclasses

import injector

from karp.foundation.commands import CommandBus
from karp.foundation.events import EventBus


@dataclasses.dataclass
class IntegrationTestContext:
    container: injector.Injector
    command_bus: CommandBus
    event_bus: EventBus



class NotFoundError(Exception):
    """Generic not found error."""

    entity_name: str = "Generic entity"

    def __init__(self, entity_id, *args, msg: str = None):
        msg = msg or f"Id: {entity_id}"
        super().__init__(f"{self.entity_name} not found. {msg}", *args)


class DiscardedEntityError(Exception):
    """Raised when an attempt is made to use a discarded Entity."""


class ConsistencyError(Exception):
    """Raised when an internal consistency problem is detected."""



import logging
import os
import pathlib

import alembic
from sqlalchemy import engine_from_config, pool, create_engine
from logging.config import fileConfig

from karp.main import config as karp_config
from karp.db_infrastructure import db

# from karp.migration_helper import entry_tables, history_tables, database_uri


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = alembic.context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)  # type: ignore
logger = logging.getLogger("alembic.env")


# def include_object(object, name, type_, reflected, compare_to):
#     return not (type_ == "table" and (name in entry_tables or name in history_tables))


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # handle testing config for migrations
    if os.environ.get("TESTING"):
        if karp_config.DATABASE_URL.drivername.startswith("sqlite"):
            db_file = pathlib.Path(karp_config.DATABASE_NAME)
            db_file.unlink(missing_ok=True)
        else:
            # connect to primary db
            default_engine = create_engine(
                karp_config.DATABASE_URL_WO_DB, isolation_level="AUTOCOMMIT"
            )
            # drop testing db if it exists and create a fresh one
            with default_engine.connect() as default_conn:
                logger.warning("dropping database: %s", karp_config.DATABASE_NAME)
                default_conn.execute(
                    f"DROP DATABASE IF EXISTS {karp_config.DATABASE_NAME}"
                )
                logger.warning("creating database: %s", karp_config.DATABASE_NAME)

                result = default_conn.execute(
                    f"CREATE DATABASE {karp_config.DATABASE_NAME}"
                )
                logger.warning("db response: %s", result)

    logger.info("migrating url: %s", karp_config.DATABASE_URL)
    connectable = config.attributes.get("connection", None)
    config.set_main_option("sqlalchemy.url", str(karp_config.DATABASE_URL))

    if connectable is None:
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),  # type: ignore
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )

    with connectable.connect() as connection:
        alembic.context.configure(
            connection=connection,
            target_metadata=db.metadata,
            compare_type=True,
        )

        with alembic.context.begin_transaction():
            alembic.context.run_migrations()


def run_migrations_offline() -> None:
    """
    Run migrations in 'offline' mode.
    """

    if os.environ.get("TESTING"):
        raise RuntimeError(
            "Running testing migrations offline currently not permitted."
        )

    alembic.context.configure(url=str(karp_config.DATABASE_URL))
    with alembic.context.begin_transaction():
        alembic.context.run_migrations()


if alembic.context.is_offline_mode():
    logger.info("Running migrations offline")
    run_migrations_offline()
else:
    logger.info("Running migrations online")
    run_migrations_online()



import abc
from typing import Iterable, Optional

import pydantic

from karp.foundation.value_objects import UniqueIdStr


class EntryRepoDto(pydantic.BaseModel):
    name: str
    entity_id: UniqueIdStr
    repository_type: str


class ListEntryRepos(abc.ABC):
    @abc.abstractmethod
    def query(self) -> Iterable[EntryRepoDto]:
        pass


class ReadOnlyEntryRepoRepositry(abc.ABC):
    @abc.abstractmethod
    def get_by_name(self, name: str) -> Optional[EntryRepoDto]:
        pass



import logging
from typing import Optional

import sqlalchemy as sa
from sqlalchemy import sql
from sqlalchemy import orm as sa_orm

from karp.foundation.events import EventBus
from karp.foundation.value_objects import UniqueId
from karp.lex.application.repositories import (
    EntryUnitOfWork,
    EntryUowRepositoryUnitOfWork,
    EntryRepositoryUnitOfWorkFactory,
    EntryUowRepository,
)

from karp.db_infrastructure.sql_unit_of_work import SqlUnitOfWork
from karp.db_infrastructure.sql_repository import SqlRepository
from karp.lex_infrastructure.sql.sql_models import EntryUowModel


logger = logging.getLogger(__name__)


class SqlEntryUowRepository(SqlRepository, EntryUowRepository):
    def __init__(
        self,
        entry_uow_factory: EntryRepositoryUnitOfWorkFactory,
        session: sa_orm.Session,
    ) -> None:
        logger.debug("create repo with session=%s", session)
        EntryUowRepository.__init__(self)  # type: ignore [arg-type]
        SqlRepository.__init__(self, session)
        self.entry_uow_factory = entry_uow_factory

    def _save(self, entry_uow: EntryUnitOfWork):
        logger.debug("saving entry_uow=%s", entry_uow)
        logger.debug("using session=%s", self._session)
        self._check_has_session()
        self._session.add(EntryUowModel.from_entity(entry_uow))
        logger.debug("session=%s, session.new=%s", self._session, self._session.new)

    def _by_id(self, id_: UniqueId, **kwargs) -> Optional[EntryUnitOfWork]:
        # stmt = sql.select(EntryUowModel).where(EntryUowModel.id == id_)
        # row = self._session.execute(stmt).first()
        stmt = (
            self._session.query(EntryUowModel)
            .filter_by(entity_id=id_)
            .order_by(EntryUowModel.last_modified.desc())
        )
        row = stmt.first()
        if row:
            return self._row_to_entity(row)
        return None

    def num_entities(self) -> int:
        self._check_has_session()
        subq = (
            self._session.query(
                EntryUowModel.entity_id,
                sa.func.max(EntryUowModel.last_modified).label("maxdate"),
            )
            .group_by(EntryUowModel.entity_id)
            .subquery("t2")
        )
        query = self._session.query(EntryUowModel).join(
            subq,
            sql.and_(
                EntryUowModel.entity_id == subq.c.entity_id,
                EntryUowModel.last_modified == subq.c.maxdate,
                EntryUowModel.discarded == False,
            ),
        )

        return query.count()

    def _row_to_entity(self, row_proxy) -> EntryUnitOfWork:
        return self.entry_uow_factory.create(
            repository_type=row_proxy.type,
            entity_id=row_proxy.entity_id,
            name=row_proxy.name,
            config=row_proxy.config,
            connection_str=row_proxy.connection_str,
            user=row_proxy.last_modified_by,
            message=row_proxy.message,
            timestamp=row_proxy.last_modified,
        )


class SqlEntryUowRepositoryUnitOfWork(SqlUnitOfWork, EntryUowRepositoryUnitOfWork):
    def __init__(
        self,
        *,
        event_bus: EventBus,
        entry_uow_factory: EntryRepositoryUnitOfWorkFactory,
        session_factory: Optional[sa_orm.sessionmaker] = None,
        session: Optional[sa_orm.Session] = None,
    ):
        SqlUnitOfWork.__init__(self, session=session)
        EntryUowRepositoryUnitOfWork.__init__(
            self,
            event_bus=event_bus,
            entry_uow_factory=entry_uow_factory,
        )
        self.session_factory = session_factory
        if self._session_is_created_here and self.session_factory is None:
            raise ValueError("Must provide one of session and session_factory")
        self._repo = None

    # def __enter__(self):
    #     logger.debug('called __enter__')
    #     if self._session_is_created_here:
    #         self._session = self.session_factory()
    #         logger.debug('created session=%s', self._session)
    #     if self._repo is None:
    #         logger.debug('creating repo')
    #         self._repo = SqlEntryUowRepository(
    #             entry_uow_factory=self.factory,
    #             session=self._session,
    #         )
    #     return super().__enter__()

    def _begin(self):
        logger.debug("called _begin")
        if self._session_is_created_here:
            self._session = self.session_factory()  # type: ignore
            logger.debug("created session=%s", self._session)
        if self._repo is None:
            self._repo = SqlEntryUowRepository(
                entry_uow_factory=self.factory,
                session=self._session,
            )
        return self

    @property
    def repo(self) -> SqlEntryUowRepository:
        if self._repo is None:
            raise RuntimeError("No entry_uow_repository")
        return self._repo



import logging
from typing import Dict

import injector
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session

from karp import lex
from karp.foundation.events import EventBus
from karp.lex.application.queries import (
    GetPublishedResources,
    GetResources,
    ResourceDto,
    ListEntryRepos,
    EntryRepoDto,
    GetReferencedEntries,
    GetEntryRepositoryId,
    EntryViews,
    GetEntryDiff,
    GetEntryHistory,
    GetHistory,
)
from karp.lex.application.queries.resources import ReadOnlyResourceRepository
from karp.lex.application.repositories import (
    EntryUowRepositoryUnitOfWork,
    EntryRepositoryUnitOfWorkFactory,
    EntryUnitOfWorkCreator,
    ResourceUnitOfWork,
)
from karp.lex_infrastructure.queries import (
    SqlGetPublishedResources,
    SqlGetResources,
    SqlListEntryRepos,
    GenericGetReferencedEntries,
    GenericGetEntryRepositoryId,
    GenericEntryViews,
    GenericGetEntryDiff,
    GenericEntryQuery,
    GenericGetEntryHistory,
    GenericGetHistory,
    SqlReadOnlyEntryRepoRepositry,
    SqlReadOnlyResourceRepository,
)
from karp.lex_infrastructure.repositories import (
    SqlEntryUowRepositoryUnitOfWork,
    SqlEntryUowV1Creator,
    SqlEntryUowV2Creator,
    SqlResourceUnitOfWork,
)


logger = logging.getLogger(__name__)


class LexInfrastructure(injector.Module):
    @injector.provider
    def get_published_resources(self, conn: Connection) -> lex.GetPublishedResources:
        return SqlGetPublishedResources(conn)

    @injector.provider
    def get_resources(self, conn: Connection) -> lex.GetResources:
        return SqlGetResources(conn)

    @injector.provider
    def read_only_resource_repo(
        self, conn: Connection
    ) -> lex.ReadOnlyResourceRepository:
        return SqlReadOnlyResourceRepository(conn)

    @injector.provider
    def list_entry_repos(self, conn: Connection) -> lex.ListEntryRepos:
        return SqlListEntryRepos(conn)

    @injector.provider
    def read_only_entry_repo_repo(
        self, conn: Connection
    ) -> lex.ReadOnlyEntryRepoRepositry:
        return SqlReadOnlyEntryRepoRepositry(conn)

    @injector.provider
    # @injector.singleton
    def entry_uow_repo(
        self,
        session: Session,
        entry_uow_factory: EntryRepositoryUnitOfWorkFactory,
        event_bus: EventBus,
    ) -> EntryUowRepositoryUnitOfWork:
        logger.debug("creating entry_repo_uow", extra={"session": session})
        return SqlEntryUowRepositoryUnitOfWork(
            session=session,
            entry_uow_factory=entry_uow_factory,
            event_bus=event_bus,
        )

    @injector.provider
    def resources_uow(
        self,
        session: Session,
        event_bus: EventBus,
    ) -> ResourceUnitOfWork:
        return SqlResourceUnitOfWork(
            session=session,
            event_bus=event_bus,
        )

    @injector.multiprovider
    def entry_uow_creator_map(self) -> Dict[str, EntryUnitOfWorkCreator]:
        return {
            "default": SqlEntryUowV2Creator,
            SqlEntryUowV1Creator.repository_type: SqlEntryUowV1Creator,
            SqlEntryUowV2Creator.repository_type: SqlEntryUowV2Creator,
        }


class GenericLexInfrastructure(injector.Module):
    @injector.provider
    def get_referenced_entries(
        self,
        resources_repo: ReadOnlyResourceRepository,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> GetReferencedEntries:
        return GenericGetReferencedEntries(
            resource_repo=resources_repo,
            entry_repo_uow=entry_repo_uow,
        )

    @injector.provider
    def gey_entry_diff(
        self,
        resources_uow: lex.ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> GetEntryDiff:
        return GenericGetEntryDiff(
            resource_uow=resources_uow,
            entry_repo_uow=entry_repo_uow,
        )

    @injector.provider
    def get_entry_repo_id(
        self,
        resources_uow: ResourceUnitOfWork,
    ) -> GetEntryRepositoryId:
        return GenericGetEntryRepositoryId(
            resource_uow=resources_uow,
        )

    @injector.provider
    def entry_views(
        self,
        get_entry_repo_id: GetEntryRepositoryId,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> EntryViews:
        return GenericEntryViews(
            get_entry_repo_id=get_entry_repo_id,
            entry_repo_uow=entry_repo_uow,
        )

    @injector.provider
    def get_history(
        self,
        resources_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> GetHistory:
        return GenericGetHistory(
            resource_uow=resources_uow,
            entry_repo_uow=entry_repo_uow,
        )

    @injector.provider
    def get_entry_history(
        self,
        resources_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> GetEntryHistory:
        return GenericGetEntryHistory(
            resource_uow=resources_uow,
            entry_repo_uow=entry_repo_uow,
        )



import uuid
from typing import Dict

from pydantic import BaseModel

from karp.foundation import events, time
from karp.foundation.value_objects import unique_id


class Event(events.Event, BaseModel):
    timestamp: float

    class Config:
        arbitrary_types_allowed = True


class AppStarted(Event):
    def __init__(self):
        self.timestamp = time.utc_now()


class ResourceCreated(Event):
    entity_id: unique_id.UniqueId
    resource_id: str
    entry_repo_id: unique_id.UniqueId
    name: str
    config: Dict
    user: str
    message: str


class ResourceLoaded(Event):
    entity_id: unique_id.UniqueId
    resource_id: str
    name: str
    config: Dict
    user: str
    message: str
    version: int


class ResourceDiscarded(Event):
    entity_id: unique_id.UniqueId
    resource_id: str
    name: str
    config: Dict
    user: str
    message: str
    version: int


class ResourcePublished(Event):
    entity_id: unique_id.UniqueId
    resource_id: str
    entry_repo_id: unique_id.UniqueId
    version: int
    name: str
    config: Dict
    user: str
    message: str


class ResourceUpdated(Event):
    entity_id: unique_id.UniqueId
    resource_id: str
    entry_repo_id: unique_id.UniqueId
    version: int
    name: str
    config: Dict
    user: str
    message: str


class EntryAdded(Event):
    entity_id: unique_id.UniqueId
    repo_id: unique_id.UniqueId
    # entry_id: str
    body: Dict
    message: str
    user: str


class EntryUpdated(Event):
    entity_id: unique_id.UniqueId
    repo_id: unique_id.UniqueId
    # entry_id: str
    body: Dict
    message: str
    user: str
    version: int


class EntryDeleted(Event):
    entity_id: unique_id.UniqueId
    repo_id: unique_id.UniqueId
    # entry_id: str
    version: int
    message: str
    user: str



from karp.foundation.value_objects import MachineName


def test_valid_machine_name_succeeds():
    name = "machine_name"
    machine_name = MachineName(name=name)
    # assert str(machine_name) == name
    assert machine_name.name == name
